/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.RandomAccessFile;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JTextField;
import javax.swing.filechooser.FileFilter;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.dialogs.BrowserDialog;
import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.Repository;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DcResources;
import net.sf.ext.jflexsett.settings.ui.ISettingsComponent;

/**
 * This field consists out of a uneditable text field and a small
 * button with a folder icon. When clicked on this button, a
 * file browser is started and the user can either select a directory
 * or file, or the user can create a new file (depending on how
 * this field is initialized)
 * 
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version 0.90
 */
public class DcFileField extends JComponent implements ISettingsComponent {

    private final JTextField text;
    private final JButton button;
    
    private ActionListener al;
    private FileFilter filter;

    private String title = DcResources.getText(Repository.labels.lblFileBrowser);
    private File file = null;

    /**
     * Creates this field
     * @param save allow saving of new files y/n
     * @param dirsOnly show directories only y/n
     */
    public DcFileField(boolean save, boolean dirsOnly) {
        this(save, dirsOnly, null);
    }
    
    /**
     * Create this field
     * @param save allow saving of new files y/n
     * @param dirsOnly show directories only y/n
     * @param filter filter for file types (example; only show mp3 files)
     */
    public DcFileField(boolean save, boolean dirsOnly, FileFilter filter) {
        this.filter = filter; 
        text = ComponentFactory.getTextField();
        text.setEditable(false);
        text.setBackground(ComponentFactory.getDisabledColor());

        button = ComponentFactory.getButton(IconLibrary._icoFileBrowser);
        
        if (save) {
            this.title = DcResources.getText(Repository.labels.lblCreateNewFile);
            button.addActionListener(new ShowSaveDialogAction());
        } else {
            if (dirsOnly) {
                this.title = DcResources.getText(Repository.labels.lblSelectDirectory);
                button.addActionListener(new ShowSelectDirectoryDialogAction());
            } else {    
                this.title = DcResources.getText(Repository.labels.lblSelectFile);
                button.addActionListener(new ShowSelectDialogAction());
            }
        }
        
        this.setLayout(Layout.getGBL());
        
        add( text,   Layout.getGBC( 0, 0, 1, 1, 80.0, 80.0
                    ,GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL,
                     new Insets(0, 0, 0, 0), 0, 0));
        button.setPreferredSize(new Dimension(25, 22));
        add( button, Layout.getGBC( 1, 0, 1, 1, 1.0, 1.0
                    ,GridBagConstraints.EAST, GridBagConstraints.NONE,
                     new Insets(0, 0, 0, 0), 0, 0));
    }
    
    /**
     * Returns the selected or newly create file or directory
     */
    public File getFile() {
        return file;
    }

    /**
     * Returns the selected or newly create file or directory
     * as a (platform dependend) string
     */
    public String getFileAsString() {
    	return file == null ? "" : file.toString();
    }
    
    /**
     * Sets the file value. The file will be set as the selected file.
     * When the file browser is started, this file is also used
     * as a starting point.
     * @param file file or directory
     */
    public void setFile(File file) {
        this.file = file; 
        text.setText(file == null ? "" : file.toString());
    }
    
    /**
     * Opens the file browser. Thru this browser a new file can be created 
     */
    protected void showFileSaveDialog() {
        BrowserDialog dialog = new BrowserDialog(title, filter);
        file = dialog.showCreateFileDialog(this, file);
        text.setText(getFileAsString());

        if (file != null) {
            try {
                RandomAccessFile random = new RandomAccessFile(file, "rw");
                random.close();
            } catch (Exception exp) {
                new MessageBox( DcResources.getText(Repository.messages.msgFileCouldNotbeCreated), 
                                MessageBox._ERROR);
                Messages.add(
                    DcResources.getText(Repository.messages.msgFileCouldNotbeCreated), 
                    Messages._ERROR);
                Messages.add(exp);
            }
        }
        dialog.dispose();
        dialog = null;
    }
    
    /**
     * Shows the file browser. Thru this browser only an existing file
     * can be selected
     */
    protected void showFileOpenDialog() {
        BrowserDialog dialog = new BrowserDialog(title, filter);
        
        file = dialog.showOpenFileDialog(this, null);
        text.setText(getFileAsString());

        if (al != null) {
        	al.actionPerformed(new ActionEvent(this.getParent(), 1, ""));
        }
        
        dialog.dispose();
        dialog = null;        
    }

    /**
     * Shows the file browser. Thru this browser only an existing 
     * directory can be selected
     */
    protected void showDirectoryOpenDialog() {
        BrowserDialog dialog = new BrowserDialog(title, filter);
        
        file = dialog.showSelectDirectoryDialog(this, null);
        text.setText(getFileAsString());
        
        dialog.dispose();
        dialog = null;        
    }

    /**
     * Add an extra listener. This listener is notified when a file has
     * been selected
     * @param al extra listener to be notified
     */
    public void addListener(ActionListener al) {
        this.al = al;
    }    
    
    private class ShowSelectDirectoryDialogAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            showDirectoryOpenDialog();
        }
    }

    private class ShowSelectDialogAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            showFileOpenDialog();
        }
    }

    private class ShowSaveDialogAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            showFileSaveDialog();
        }
    }

	public void setValue(Object value) {
        if (value != null && value.toString().length() > 0) {
        	File file = new File(value.toString());
        	setFile(file);
        } else {
        	setFile(null);
        }
	}

	public Object getValue() {
		return file == null ? "" : file.toString();
	}
}
