/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components.panels;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.Collection;

import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import javax.swing.ToolTipManager;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.MainFrame;
import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.console.dialogs.QuestionBox;
import net.sf.dc.console.dialogs.SaveQuestionBox;
import net.sf.dc.console.reporting.ReportingDialog;
import net.sf.dc.console.software.SoftwareCategoryForm;
import net.sf.dc.core.Repository;
import net.sf.dc.core.actions.ToggleQuickViewAction;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.exceptions.ValidationException;
import net.sf.dc.processes.DataTask;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.wf.requests.IRequest;
import net.sf.dc.wf.requests.RemoveFromCacheRequest;
import net.sf.dc.wf.requests.RemoveRowsRequest;
import net.sf.dc.wf.requests.StatusUpdateRequest;

/**
 * Base for all search panels (for objects like Software, Mp3 and Movie).
 * Has an implemenation for supporting a child panel (in case of an Object
 * with children). The functionality is not fully implemented in this class
 * as only (for now) the Audio Track and Audio CD panel have a parent-child
 * relation. 
 *
 * @author Robert Jan van der Waals
 * @since 1.4
 * @version 1.9
 */
public abstract class TableSearchPanel extends TablePanel {

    protected TableSearchPanel childPanel;
    private QuickViewPanel quickView;
    
    private String parentID = "";
    
    /**
     * Initializes the panel. Should be called by the child by calling
     * super(module). The popup menu and the action panel are automatically
     * created.
     * @param module the module for which this panel is created
     * @param title the title of this panel
     * @param icon the icon of this panel
     */
    public TableSearchPanel(DCModule module, String title, ImageIcon icon) {
        super(module, title, icon);
        setHelpIndex("dc.Searching.SearchResult");

        boolean imdb = getModule().isImdbUpdateAllowed();
        boolean amazon = getModule().isAmazonUpdateAllowed();
        boolean open = this instanceof SoftwareCategoryForm.CatagorySearchPanel ?
                       false : true;

        JComponent[] extraComponents = getAllExtraComponents();
        
        panelTableActions = new TableActionsPanel(false, extraComponents);
        panelTableActions.addSaveActionListener(new SaveUpdatedValuesAction());
    
        popupMenu = new PopupMenu(imdb, amazon, true, false, open, true);
        popupMenu.addAmazonActionListener(new AmazonUpdateAction());
        popupMenu.addIMDBActionListener(new ImdbUpdateAction());
        popupMenu.addSaveActionListener(new SaveSelectedValuesAction());
        popupMenu.addDeleteActionListener(new DeleteValuesAction());
        
        JMenuItem menuQuickView = new JMenuItem(DcResources.getText(Repository.labels.lblToggleQuickView));
        menuQuickView.addActionListener(new ToggleQuickViewAction());
        popupMenu.add(menuQuickView);
        
        buildPanel();
    }
    
    public void showQuickView(boolean b) {
        quickView.setVisible(b);
    }
    
    /**
     * Retrieves the module specific components 
     */
    private JComponent[] getAllExtraComponents() {
        JComponent[] components = new JComponent[0];
        
        JComponent[] childComponents = new JComponent[0];
        if (hasChild()) {
        	childComponents = getChildModule().getSearchPanel().getExtraComponents();
        }
        
        JComponent[] parentComponents = getExtraComponents();
        components = new JComponent[parentComponents.length + childComponents.length];
        int counter = 0; 
        
        if (childComponents != null) {
            for (int i = 0; i < childComponents.length; i++) {
                components[counter] = childComponents[i];
                counter++;
            }
        }
        
        if (parentComponents != null) {
            for (int i = 0; i < parentComponents.length; i++) {
                components[counter] = parentComponents[i];
                counter++;
            }
        }
        return components;
    }

    public abstract JComponent[] getExtraComponents();
    
    /**
     * When this search panel is acting as a child for another panel the ID
     * of the parent item for which items are shown in the child panel
     * is stored. 
     * @param sID parent ID
     */
    public void setParentID(String sID) {
        parentID = sID;
    }
    
    /**
     * Returns the parent ID for which items are shown (in case this is
     * a child panel)
     */
    public String getParentID() {
    	return parentID;
    }
    
    public void checkForChanges() {
        cancelEdit();
        
        boolean saved = changesSaved();
        saved = hasChild() ? saved && childPanel.changesSaved() : saved;
        if (!saved) {
            new QuestionBox(DcResources.getText(Repository.messages.msgNotSaved));
            if (QuestionBox.result.equals(Boolean.TRUE)) {
                saveChangedValues(false);
                
                if (hasChild()) {
                    childPanel.saveChangedValues(false);
                }
            }
        }        
    }
    
    /**
     * Clears this table and the child table
     */
    public void clearTable() {
        checkForChanges();
        
    	super.clearTable();
        
        if (hasChild()) {
        	childPanel.clearTable();
            childPanel.setParentID(null);
        }
        
        if (quickView != null) {
            quickView.clear();
        }
    }
    
    /**
     * Removes the object with the given ID from the cache
     * @param sID Object ID
     */
    public void removeFromCache(String sID) {
    	table.removeFromCache(sID);
    }
    
    public DcObject getSelectedObject() {
    	int row = table.getSelectedRow();
    	DcObject dco = null;
    	if (row != -1) {
    		dco = getObjectForRow(row, true);
    	} else {
    		new MessageBox(DcResources.getText(Repository.messages.msgNoRowsSelected),
                    	   MessageBox._WARNING);
    	}
    	return dco;
    }
    
    public DCModule getModuleForRow(int row) {
        return table.getModuleForRow(row);
    }
    
    /**
     * Opens the item form
     */
    public void openItemForm() {
        cancelEdit();
        int row = table.getSelectedRow();
        if (row == -1) {
            new MessageBox(DcResources.getText(Repository.messages.msgSelectRowToOpen),
                           MessageBox._WARNING);
        } else {
            DcObject dco = getObjectForRow(row, true);
            openItemForm(dco, true);
        }
    }
    
    public void updateRow(String id, DcObject dco, boolean overwrite, boolean allowDeletes, boolean mark) {
        super.updateRow(id, dco, overwrite, allowDeletes, mark);
        if (quickView != null)
            quickView.setObject(getObjectForRow(table.getSelectedRow(), true));
    }    
    
    /**
     * Creates a PDF report. All visible fields are reported on.
     * A file browser is opened in which the user can indicate where
     * the pdf file should be stored.
     */
    public void createPdfReport() {
        DcObject[] objects = new DcObject[table.getRowCount()];
        for (int i = 0; i < objects.length; i++) {
        	objects[i] = getObjectForRow(i, true); 
        }
        
        if (objects.length > 0) {
            ReportingDialog dialog = new ReportingDialog(objects);
            dialog.setVisible(true);
        } else {
            new MessageBox(DcResources.getText(Repository.messages.msgReportNoItems), 
                           MessageBox._WARNING);
        }
    }     
    
    public abstract void removeNewChild (String sID);
    
    
    /**
     * Permanently removes an item / items from the database.
     * The user must always confirm the delete action first before 
     * it is executed !
     */
    public void deleteValues() {
        if (!isTaskRunning()) {
            int[] rows = table.getSelectedRows();
            if (rows.length > 0) {
                task = new DeletingThread(rows, getClass());
                task.start();
            } else {
                new MessageBox(DcResources.getText(Repository.messages.msgSelectItemToDel), MessageBox._WARNING);    
            }
            
        } else {
            new MessageBox(DcResources.getText(Repository.messages.msgJobRunning),
            		       MessageBox._WARNING);    
        }
    }    

    /**
     * Saves the selected values
     */
    public void saveSelectedValues() {
        if (!isTaskRunning()) {
            int[] rows = table.getSelectedRows();
            task = new SavingThread(rows, getClass());
            task.start();
        } else {
            new MessageBox(DcResources.getText(Repository.messages.msgJobRunning),
                           MessageBox._WARNING);    
        }
    }
    
    /**
     * Saves all the changed values. The orginal values are stored 
     * in the 'cache' hashtable.
     * @param threaded indicates if the save should be done threaded or in the
     *                 ui thread
     */
    public void saveChangedValues(boolean threaded) {
        if (!isTaskRunning()) {
            int[] rows = table.getChangedObjectsIndices();
            
            boolean childrenSaved = true;
            if (hasChild()) {
                childPanel.cancelEdit();
            	childrenSaved = childPanel.changesSaved();
            }
            
            if (rows.length > 0 || !childrenSaved) {
                task = new SavingThread(rows, getClass());
                if (threaded) {
                    task.start();   
                } else {
                	task.run();
                }
            } else {
                new MessageBox(DcResources.getText(Repository.messages.msgNoChangesToSave), MessageBox._WARNING);    
            }
        } else {
            new MessageBox(DcResources.getText(Repository.messages.msgJobRunning),
            		       MessageBox._WARNING);    
        }
    }
    
    /**
     * Returns a data crow object for the selected row.
     * Child objects are added to the parent data crow object. 
     * @param row the row for which a data crow object should be created
     */
    public DcObject getObjectForRow(int row, boolean allChildren) {
        DcObject dco = table.getDCObject(row);
        dco.removeChildren();
        if (hasChild()) {
            DcObject[] children = null;
            if (allChildren) {
                children = childPanel.getObjects();
            } else {
            	children = childPanel.getChangedObjects();
            }
            
            for (int j = 0; j < children.length; j++) {
                DcObject child = children[j];
                child.addRequest(new RemoveFromCacheRequest(childPanel.getModule().getIndex(), child.getID()));
                dco.addChild(child);
            }
        }
        
        dco.setSilence(true);
        dco.setBatch(true);        
        return dco;
    }  
    
    public void removeRows(String[] ids) {
        if (hasChild()) {
            childPanel.clearTable();
            childPanel.setParentID(null);
        }
        
        if (quickView != null) {
            quickView.clear();
        }
        
        super.removeRows(ids);
    }
    
    /**
     * Retrieves all changed objects
     * Note: use carefully. In case of an update all this might mean that thousands 
     * of items are loaded into memory ! Right now this method is solely used for child panels
     */
    public DcObject[] getChangedObjects() {
        return table.getChangedObjects();
    }
    
    /**
     * Indicates whether all changes to the UI table have been saved
     * to the database
     */
    public boolean changesSaved() {
        return table.noChanges();
    } 

    /**
     * Retrieves children for the selected parent object. The children are retrieved
     * by the ID of the parent. 
     */
    public void saveChildItems() {
        if (hasChild() && table.getSelectedRow() > -1) {
            childPanel.cancelEdit();
            if (!childPanel.changesSaved()) {
            	new QuestionBox(DcResources.getText(Repository.messages.msgNotSaved));
                if (QuestionBox.result.equals(Boolean.TRUE)) {
                    saveChangedValues(false);
                }
            }
        }
    }
    
    public abstract void loadChildValues();
    
    public void setBackground(Color color) {
        super.setBackground(color);
    }    
    
    public void addObjects(Collection collection) {
    	table.addObjects(collection);
    }

    /**
     * Initializes the panel
     */
    private void buildPanel() {
        JPanel panelStatus = getStatusPanel();
        JPanel panelResult = new JPanel(false);

        //**********************************************************
        //Search result panel
        //**********************************************************
        panelResult.setLayout(Layout.getGBL());

        table = ComponentFactory.getDCTable(getModule(), false, true);
        table.addMouseListener(tableMouseListener);

        JScrollPane scrollResults = new JScrollPane(table);
        scrollResults.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        scrollResults.getViewport().setScrollMode(JViewport.SIMPLE_SCROLL_MODE);
        
        panelResult.add(scrollResults,  Layout.getGBC( 0, 0, 1, 1, 80.0, 80.0
                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                        new Insets(5, 5, 5, 5), 0, 0));
        
        int length = 1;
        
        table.getSelectionModel().addListSelectionListener(new RowSelectionListener());
        table.addKeyListener(new ChangeRowByKey());
        
        if (hasChild()) {
        	DCModule module  = getChildModule();
            childPanel = module.getSearchPanel();
            
            JScrollPane scrollResults2 = new JScrollPane(childPanel.table);
            scrollResults2.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
            scrollResults2.getViewport().setScrollMode(JViewport.SIMPLE_SCROLL_MODE);            
            
            panelResult.add(    scrollResults2,  Layout.getGBC( 0, 1, 1, 1, 80.0, 80.0
                               ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                                new Insets(5, 5, 5, 5), 0, 0));
            length = 2;
        }
        
        quickView = getModule().getQuickView();
        if (quickView != null) { 
            boolean showQuickView = DcSettings.getValueAsBoolean(Repository.settings.stShowQuickView);
            quickView.setVisible(showQuickView);
            panelResult.add(quickView,  Layout.getGBC( 1, 0, 1, length, 1.0, 1.0
                           ,GridBagConstraints.NORTHWEST, GridBagConstraints.VERTICAL,
                            new Insets(5, 5, 5, 0), 0, 0));
        }
        
        
        //**********************************************************
        //Main panel
        //**********************************************************
        setLayout(Layout.getGBL());
        
        add(    panelResult,      Layout.getGBC( 0, 3, 3, 1, 100.0, 100.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                new Insets(5, 5, 5, 5), 0, 0)); 
        add(    panelTableActions,Layout.getGBC( 0, 4, 3, 1, 1.0, 1.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                new Insets(5, 5, 5, 5), 0, 0));
        add(    panelStatus,      Layout.getGBC( 0, 5, 3, 1, 1.0, 1.0
                ,GridBagConstraints.SOUTH, GridBagConstraints.HORIZONTAL,
                new Insets(0, 5, 0, 5), 0, 0));

        ToolTipManager.sharedInstance().registerComponent(table);
    }
   
    
    
    // ****************************************************************************
    // ACTIONS
    // ****************************************************************************
    private class DeletingThread extends DataTask {

        public DeletingThread(int[] rows, Class clazz) {
        	super(rows);
        }
        
        public void run() {
            try {
                updateProgressBar(0);
                initProgressBar(rows.length);   
                
                table.setCursor(new Cursor(Cursor.WAIT_CURSOR));
                
                setStatus("Deleting " + rows.length + " items from to the database");
                
                startRunning();
                cancelEdit();
                denyActions();
                
                int[] rows = table.getSelectedRows();
    
                if (rows == null || rows.length <= 0) {
                    new MessageBox(DcResources.getText(Repository.messages.msgNoRowsSelected),
                                   MessageBox._WARNING);
                    stopRunning();
                } else {
                    new QuestionBox(DcResources.getText(Repository.messages.msgDeleteQuestion));
                    if (!QuestionBox.result.booleanValue()) {
                        stopRunning();
                    }
                }
    
                int counter = 1;
                for (int i = rows.length -1; i > -1 && keepOnRunning(); i--) {
                    updateProgressBar(counter++);
                    
                    DcObject dco = getObjectForRow(rows[i], false);
                    dco.addRequest(
                            new RemoveRowsRequest(getModule().getIndex(), 
                                                  MainFrame._SEARCHTAB,
                                                  new String[] {dco.getID()}));
                    dco.clearValues();
                    dco.markAsUnchanged();
                    
                    dco.setValue(DcObject._ID, table.getObjectID(rows[i]));
                    dco.setSilence(true);
                    dco.delete(); 
                    
                    if (i == 0) {
                    	dco.setEndOfBatch(true);
                        IRequest request = 
                            new StatusUpdateRequest(getModuleForRow(rows[i]).getIndex(),
                            		                MainFrame._SEARCHTAB,
												    Repository.messages.msgDeleteSuccessfull);
                        request.setExecuteOnFail(true);
                        dco.addRequest(request);
                    } else {
                        dco.setEndOfBatch(false);
                    }
                    
                    try {
                    	sleep(300);
                    } catch (Exception ignore) {} // doesn't want to sleep..
                }
                
                table.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
            } finally {
                table.setSelectedRow(0);
            	stopRunning();
            	allowActions();
            }
        }
    }    
    
    private class SavingThread extends DataTask {
        public SavingThread(int[] rows, Class clazz) {
            super(rows);
        }
        
        public void run() {
            try {
                updateProgressBar(0);
                initProgressBar(rows.length);
                setStatus("Saving " + rows.length + " items to the database");
                table.setCursor(new Cursor(Cursor.WAIT_CURSOR));
                
                startRunning();
                cancelEdit();
                denyActions();
                
                if (hasChild() && !childPanel.changesSaved() && ((rows == null || rows.length == 0))) {
                	rows = table.getSelectedRows();
                    childPanel.cancelEdit();
                } 
                
                if (rows == null || rows.length == 0) {
                    stopRunning();
                }
                
                boolean ignoreErrors = false;
                int counter = 1; 
                for (int i = rows.length - 1; i > -1 && keepOnRunning(); i--) {
                	try {
	                    updateProgressBar(counter++);
	                    DcObject dco = getObjectForRow(rows[i], false);
	                    dco.setEndOfBatch(i == 0);
	                    dco.addRequest(new RemoveFromCacheRequest(getModule().getIndex(), dco.getID()));
                        dco.saveUpdate();
                    } catch (ValidationException vExp) {
                        if (!ignoreErrors) {
                            SaveQuestionBox question = new SaveQuestionBox(vExp);
                            if (question.getResult() == SaveQuestionBox._CANCEL) {
                                stopRunning();
                            } else if (question.getResult() == SaveQuestionBox._IGNORE) {
                                ignoreErrors = true;
                            }
                        }
                    }
                    
                    try {
                        sleep(300);
                    } catch (Exception ignore) {} // doesn't want to sleep..
                }
                setStatus(DcResources.getText(Repository.messages.msgSaveHasBeenExecuted));
            } finally {
                table.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
                allowActions();
            	stopRunning();
            }
            
            if (hasChild()) {
            	childPanel.saveChangedValues(false);
            }
        }
    }    
    
    private class RowSelectionListener implements ListSelectionListener {
    	public void valueChanged(ListSelectionEvent e){
            if (quickView != null && quickView.isVisible()) {
                int row = table.getSelectedRow();
                if (row > -1) 
                    quickView.setObject(getObjectForRow(row, true));
            }
            
            if (hasChild()) {
        	   saveChildItems();
        	   loadChildValues();
            }
    	}
    }
    
    private class ChangeRowByKey implements KeyListener {
        private int rowFrom = -1;
        private int rowTo = -1;
        
        public void keyPressed(KeyEvent e) {
           if (e.getKeyCode() == KeyEvent.VK_UP || e.getKeyCode() == KeyEvent.VK_DOWN) {
               rowFrom = table.getSelectedRow();
           }
        }
        
        public void keyReleased(KeyEvent e) {
            if (e.getKeyCode() == KeyEvent.VK_UP || e.getKeyCode() == KeyEvent.VK_DOWN) {
                rowTo = table.getSelectedRow();
                if (rowFrom != rowTo) {
                    if (hasChild()) {
                        saveChildItems();
                        loadChildValues();
                    }
                    
                    if (quickView != null && quickView.isVisible()) {
                        quickView.setObject(getObjectForRow(rowTo, true));
                    }
                }
            }
        }
        
        public void keyTyped(KeyEvent arg0) {}
    }    
    
    private class SaveSelectedValuesAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            saveSelectedValues();
        }
    }

    private class AmazonUpdateAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            startAmazonUpdate();
        }
    }

    private class SaveUpdatedValuesAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            saveChangedValues(true);
        }
    }

    private class DeleteValuesAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            deleteValues();
        }
    }

    private class ImdbUpdateAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            startIMDBUpdate();
        }
    }
}
