/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components.tables;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableColumn;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.components.DcNumberField;
import net.sf.dc.console.components.DcRatingComboBox;
import net.sf.dc.console.components.renderers.ComboBoxTableCellRenderer;
import net.sf.dc.console.components.renderers.DcTableCellRenderer;
import net.sf.dc.console.components.renderers.DcTableHeaderRenderer;
import net.sf.dc.console.components.renderers.DcTableHeaderRendererRequired;
import net.sf.dc.console.components.renderers.DcTableRowIndexCellRenderer;
import net.sf.dc.console.components.renderers.ModuleTableCellRenderer;
import net.sf.dc.console.components.renderers.NumberTableCellRenderer;
import net.sf.dc.console.components.renderers.PictureTableCellRenderer;
import net.sf.dc.console.components.renderers.RatingTableCellRenderer;
import net.sf.dc.console.components.renderers.TimeFieldTableCellRenderer;
import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.console.dialogs.UpdateAllDialog;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.generic.GenericObject;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.settings.definitions.DcFieldDefinition;
import net.sf.dc.settings.definitions.Definitions;
import net.sf.ext.util.StringUtils;

/**
 * Table implementation with caching of changed values.
 * Can either be initialized as a simple table or a table for module (with
 * all the correct components and headers)
 *
 * @author Robert Jan van der Waals
 * @since 1.4
 * @version 1.9
 */
public class DcTable extends JTable  {
    
    private final DCModule module;
    private final DcTableModel tableModel;
    private final Hashtable cache = new Hashtable();
    
    private final TableValueChangedAction tableChangeListener = 
        new TableValueChangedAction();    
    
    private final boolean caching;
    private final boolean readonly;
    private boolean ignoreEdit = false;
    
    private ArrayList columnsHidden = new ArrayList();
    private Map columns = new HashMap();
    private TableColumn rowIndexColumn;

    /**
     * Initializes a very simple table
     */
    public DcTable(boolean readonly, boolean caching) {
        super(new DcTableModel());
        this.readonly = readonly;
        createRowIndexColumn();
        
        module = null;
        
        addSystemColumns();
        setProperties();
        
        tableModel = (DcTableModel) getModel();
        this.caching = caching;
        readonly = false;
    }

    /**
     * Initializes a table for a specific module
     * @param module the mocule for which this table is created
     * @param readonly allow / block changes 
     */
    public DcTable(DCModule module, boolean readonly, boolean caching) {
        super(new DcTableModel());
        tableModel = (DcTableModel) getModel();

        this.caching = caching;
        this.module = module;
        this.readonly = readonly;
        
        buildTable();
        setListeningForChanges(true);
    }

    private int getFieldForColumnIndex(int columnIndex) {
        int field = -1;
        for (Iterator iter = columns.values().iterator(); iter.hasNext(); ) {
            TableColumn column = (TableColumn) iter.next();
            if (column.getModelIndex() == columnIndex) {
                Integer identifier = (Integer) column.getIdentifier();
                
                if (identifier != null)
                    field = identifier.intValue();
            }
        }
        return field;
    }    
    
    public int getColumnIndexForField(int field) {
        int columnIndex = -1;
        for (Iterator iter = columns.values().iterator(); iter.hasNext(); ) {
            TableColumn column = (TableColumn) iter.next();
            Integer identifier = (Integer) column.getIdentifier();
            if (identifier != null && identifier.intValue() == field) {
                columnIndex = column.getModelIndex();
            }
        }
        return columnIndex;
    }
    
    private int addRow() {
        Object[] row = new Object[module.getFieldCount()];
        
        DcObject dco = module.getDCObject();
        int[] fields = dco.getFieldIndices();
        
        for (int i = 0; i < fields.length; i++) {
            int field = fields[i]; 
            int col = getColumnIndexForField(field);
            row[col] = dco.getValue(field);
        }
        tableModel.addRow(row);

        return (getRowCount() - 1);
    }
    
    public void addObject(DcObject dco) {
        addObject(dco, true);
    }
    
    public void addObject(DcObject dco, boolean setSelected) {
        int[] fields = dco.getFieldIndices();
        
        int row = addRow();
        for (int i = 0; i < fields.length; i++) {
            int field = fields[i];
            int col = getColumnIndexForField(field);
            Object value = dco.getValue(fields[i]);
            tableModel.setValueAt(value, row, col);
        }
        
        if (module.getIndex() == DCModules._GENERIC) {
            //dco.setValue(GenericObject._SYS_MODULE, dco.getModule());
            int col = getColumnIndexForField(GenericObject._SYS_MODULE);
            Object value = dco.getModule();
            tableModel.setValueAt(value, row, col);
        }
        
        if (setSelected) {
        	setSelectedRow(getRowCount() - 1);
        }
    }
    
    public void addObjects(Collection objects) {
        for (Iterator iter = objects.iterator(); iter.hasNext();) {
            DcObject dco = (DcObject) iter.next();
            addObject(dco, false);
        }
        setSelectedRow(getRowCount() - 1);
    }
    
    /**
     * Updates the row with the given ID with the data (values)
     * @param ID the ID (Product Object ID)
     * @param values the new values
     */
    public void updateRow(String ID, DcObject dco) {
        cancelEdit();

        int[] fields = dco.getFieldIndices();
        int row = getRowNumberWithID(ID);
        
        if (row != -1) {
            for (int i = 0; i < fields.length; i++) {
                int field = fields[i];
                int column = getColumnIndexForField(field);
                Object value = dco.getValue(field);

                tableModel.setValueAt(value, row, column);
                addRowToCache(row, column);
            }
        }
    }     
    
    /**
     * Indicates if the table data can be edited
     */
    public boolean isReadOnly() {
    	return readonly;
    }
    
    /** 
     * Indicates if changes have been made to the data of the table
     */
    public boolean noChanges() {
        return cache.size() == 0;
    }    
    
    /**
     * Ignores the current edit action of a cell
     * @param b
     */
    public void ignoreEdit(boolean b) {
    	ignoreEdit = b;
    }
    
    /**
     * (Re-) applies all headers
     */
    public void applyHeaders() {
    	removeSystemColumns();
        for (Enumeration enumCols = getColumnModel().getColumns(); enumCols.hasMoreElements(); ) {
            TableColumn column = (TableColumn) enumCols.nextElement();
            column.setHeaderRenderer(getTableHeaderRenderer());
            columns.put(column.getIdentifier(), column);
        }
        addSystemColumns();
    }
    
    public void moveRowDown() {
        int row = getSelectedRow();
        
        if (row > -1) {
            int total = getRowCount();
            if (row < total -1) {
                int destination = row + 1;
                tableModel.moveRow(row, row, destination);
                setSelectedRow(destination);
            }
        } else {
            new MessageBox("No row selected. First select a row to move", MessageBox._WARNING);
        }
    }
    
    public void moveRowUp() {
        int row = getSelectedRow();
        
        if (row > -1) {
            if (row != 0) {
                int destination = row - 1;
                tableModel.moveRow(row, row, destination);
                setSelectedRow(destination);
            }
        } else {
            new MessageBox("No row selected. First select a row to move", MessageBox._WARNING);
        }
    }

    /**
     * All columns to be removed are still visible in the table. The columns 
     * are retrieved, stored in a Vector and then removed from the view
     * @param columns The model indices of the columns
     */
    public void removeFields(int[] fields) {
        cancelEdit();
        removeSystemColumns();
        
        for (int i = 0; i < fields.length; i++) {
        	int columnIndex = getColumnIndexForField(fields[i]);
        	
            int tableIndex = convertColumnIndexToView(columnIndex);
            try {
                TableColumn column = getColumnModel().getColumn(tableIndex);
                columnsHidden.add(column); 
                removeColumn(column);
            } catch (Exception ignore) {}
        }
        
        addSystemColumns();
    }
    
    /**
     * Clears the cached objects
     */
    public void clearCache() {
    	cache.clear();
    }
    
    /**
     * Removes the object with the given ID from the cache (unmark as changed)
     * @param sID object ID
     */
    public void removeFromCache(String sID) {
    	cache.remove(sID);
    }
    
    /**
     * Returns all Data Crow Objects currently persisting in the (UI) table
     */
    public DcObject[] getDCObjects() {
        DcObject[] objects = new DcObject[tableModel.getRowCount()];
        for (int i = 0; i < objects.length; i++) {
            objects[i] = getDCObject(i);
        }
        return objects;
    }   
    
    private boolean isCached(String id) {
    	return cache.containsKey(id);
    }
    
    public DcObject getDCObject(int row) {
        int col = getColumnIndexForField(DcObject._ID);
        String id = (String) getValueAt(row, col, true);
        boolean isCached = isCached(id);
        
        DcObject dco;
        if (isCached) {
        	dco = (DcObject) cache.get(id);
        } else { 
        	dco = getObjectForRow(row);
        	dco.markAsUnchanged();
        }
        return dco;
    }
    
    private DcObject getObjectForRow(int row) {
        cancelEdit();
        DcObject dco = null;
        
        if (row == -1) {
        	new MessageBox(DcResources.getText(Repository.messages.msgNoRowsSelected), 
                           MessageBox._WARNING);
        } else {
            dco = getModuleForRow(row).getDCObject();
            for (Iterator iter = columns.values().iterator(); iter.hasNext(); ) {
                TableColumn column = (TableColumn) iter.next();

                int col = column.getModelIndex();
                int field = getFieldForColumnIndex(col);
                Object value = getValueAt(row, col, true);
                dco.setValue(field, value);
            }
        }
        return dco;
    }
    
    public DCModule getModuleForRow(int row) {
        if (module.getIndex() == DCModules._GENERIC) { 
            int col = getColumnIndexForField(GenericObject._SYS_MODULE);
            Object value = getValueAt(row, col, true);
            if (value instanceof DCModule) { 
                return (DCModule) value;
            } else { 
                String s = value.toString();
                for (Iterator iter = DCModules.getModules(); iter.hasNext(); ) {
                    DCModule mod = (DCModule) iter.next();
                    if (mod.getName().equals(s))
                        return mod;
                }
                return module;
            }
        } else {
            return module;
        }
    }
    
    public String getObjectID(int row) {
        int col = getColumnIndexForField(DcObject._ID);
        return (String) getValueAt(row, col, true);
    }
    
    /**
     * Retrieve all changed objects. Use with care, this could be heavy on the
     * system when many rows have been changed. Consider using the method
     * getChangedObjectsIndices
     */
    public DcObject[] getChangedObjects() {
        DcObject[] objects = new DcObject[cache.size()];
        int counter = 0;
        for (Enumeration enumKeys = cache.keys(); enumKeys.hasMoreElements(); counter++) {   
            String sID = enumKeys.nextElement().toString();
            DcObject obj = (DcObject) cache.get(sID);
            objects[counter] = obj;
        }
        return objects;
    }    
    
    /**
     * Retrieves the rownumbers of all changed objects
     */
    public int[] getChangedObjectsIndices() {
        cancelEdit();
        int[] rows = new int[cache.size()];
        int counter = 0;
    	for (Enumeration enumKeys = cache.keys(); enumKeys.hasMoreElements(); counter++) {   
            String sID = enumKeys.nextElement().toString();
            int row = getRowNumberWithID(sID);
            rows[counter] = row;
        }
        return rows;
    }

    /**
     * Gets a value from the model (hidden or visible) 
     */
    public Object getValueAt(int row, int col, boolean hidden) {
    	return hidden ? tableModel.getValueAt(row, col) : super.getValueAt(row, col);
    }
    
    /**
     * Clear the table of all values. Clear the cache as well.
     */
    public void clearTable() {
        cache.clear();
        cancelEdit();

        setListeningForChanges(false);
        tableModel.setRowCount(0);
        setListeningForChanges(true);
    }
    
    /**
     * Returns the row number of the item with the given ID
     * (product object ID; ie. the value of Software._ID)
     * @param ID
     */
    public int getRowNumberWithID(String ID) {
        cancelEdit();
        for (int i = 0; i < tableModel.getRowCount(); i++) {
            if (ID.equals(getObjectID(i))) 
                return i;
        }
        return -1;
    }  
    
    /**
     * Sets the amount of columns to be displayed
     * @param count columns amount
     */
    public void setColumnCount(int count) {
    	tableModel.setColumnCount(count);
    }
    
    /**
     * Remove the rows from the table with the given indices
     */
    public void removeRows(int[] rows) {
        cancelEdit();
        for (int i = rows.length -1; i > -1; i--) {
        	if (caching) {
                int row = rows[i];
                int col = getColumnIndexForField(DcObject._ID);
    			removeFromCache((String) getValueAt(row, col, true));
        	}
            tableModel.removeRow(rows[i]);
        }
    }   
    
    /**
     * Removes a row from the UI table
     * @param row row index
     */
    public void removeRow(int row) {
        tableModel.removeRow(row);
    }    
    
    /**
     * Sets the amount of rows to be displayed
     * @param count rows amount
     */
    public void setRowCount(int count) {
        tableModel.setRowCount(count);
    }    

    /**
     * Tries to programmatically set the selected row in the table.
     * @param row index
     */
    public void setSelectedRow(int row) {
        try {
            if (getSelectedRow() > -1) {
                removeColumnSelectionInterval(0, getColumnCount() -1);
                removeRowSelectionInterval(getSelectedRow(), getSelectedRow());
            }
                
            addRowSelectionInterval(row, row);
            addColumnSelectionInterval(0, getColumnCount() -1);
        } catch (Exception ignore) {}
    }
    
    /**
     * Updates a row at the given rownumber. If indicated this will
     * update existing data.
     */
    public void updateRowAt(int row,
                            DcObject dco,
                            boolean overwrite,
                            boolean allowDeletes,
                            boolean mark) {
        cancelEdit();

        if (!mark) {
            setListeningForChanges(false);
            removeFromCache(dco.getID());
        }
        
        int[] indices = dco.getFieldIndices();
        for (int i = 0; i < indices.length; i++) {
            try {
                TableColumn column = (TableColumn) columns.get(new Integer(indices[i]));
                int col = column.getModelIndex();
                
                Object newValue = dco.getValue(indices[i]);
                Object existingValue = tableModel.getValueAt(row, col);
                
                if (overwrite && ((newValue != null && !newValue.equals("")) || allowDeletes)) {
                    tableModel.setValueAt(newValue, row, col);
                } else if (existingValue == null || existingValue.equals("")) {
                    tableModel.setValueAt(newValue, row, col);
                }
            } catch (Exception ignore) {}
        }
        
        if (!mark) 
            setListeningForChanges(true);   
    } 
    
    public void addRow(Object[] row) {
    	tableModel.addRow(row);
    }
    
    /**
     * Add a changed row to the cache. Can also be used to force save items
     * (add a unchanged row to the cache) 
     * @param row changed row number
     * @param column changed column number
     */
    public void addRowToCache(int row, int column) {
        setListeningForChanges(false);
        
        if (row != -1 && column != -1) {
            int col = getColumnIndexForField(DcObject._ID);
            String id = getValueAt(row, col, true).toString();
            
            if (id != null && !id.equals("")) {
                DcObject dco;
                if (!cache.containsKey(id)) {
                    dco = getObjectForRow(row);
                    dco.markAsUnchanged();
                } else {
                    dco = (DcObject) cache.get(id);
                }
                
                int field = getFieldForColumnIndex(column);
                dco.setValue(field, tableModel.getValueAt(row, column));
                cache.put(id, dco);
            }
        }        
        setListeningForChanges(true);
    }      
    
    /**
     * Fills the table with the specified values
     * @param data list containing the rows of data to be displayed
     */
    public void fillTable(Collection objects) {
        clearTable();

        setListeningForChanges(false);
        
        if (objects != null) {
            for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
                DcObject dco = (DcObject) iter.next(); 
            	addObject(dco, false);
            }
        }
        
        setSelectedRow(getRowCount() -1);
        setListeningForChanges(true);
    }    
    
    /**
     * Tries to apply the current editing of the table cell.
     * If the current value cannot be applied, the editing is simply
     * cancelled (changes are discarded) 
     */
    public void cancelEdit() {
        int selectedRow = getSelectedRow();
        if (selectedRow != -1) {
            for (int i = 0; i < getColumnCount(); i++) {
                boolean success = false;
                try {
                    getCellEditor(selectedRow, i).stopCellEditing();
                    success = true;
                } catch (Exception ignore) {}
                
                if (!success) {
                    getCellEditor(selectedRow, i).cancelCellEditing();
                }
            }
        }
    }   

    //*************************************************************************
    //Dialogs
    //*************************************************************************
    
    /**
     * Opens the Update All dialog
     */
    public void openUpdateAllDialog() {
        UpdateAllDialog dialog = new UpdateAllDialog(module.getDCObject(), DataCrow.mainFrame);
        // if the update was a success
        if (dialog.isSuccess()) {
            // retrieve the data which will be applied on the table
            
            
            DcObject dco = dialog.getValues();
            setListeningForChanges(false);
            if (dialog.isUpdateSelectedItemsOnly()) {
            	int[] rows = getSelectedRows();
                int[] fields = dco.getFieldIndices();
                for (int i = 0; i < rows.length; i++) {
                    int row = rows[i];
                    for (int j = 0; j < fields.length; j++) {
                        // only update a field with a 'real' / not empty value
                        Object value = dco.getValue(fields[j]);
                        if (!isEmptyValue(value)) {
                            int col = getColumnIndexForField(fields[i]);
                            tableModel.setValueAt(value, row, col);
                            addRowToCache(row, col);
                        }
                    }
                }
            } else {
                int rows = tableModel.getRowCount();
                int[] fields = dco.getFieldIndices();
                for (int i = 0; i < rows; i++) {
                    int row = i;
                    for (int j = 0; j < fields.length; j++) {
                        // only update a field with a 'real' / not empty value
                        Object value = dco.getValue(fields[j]);
                        if (!isEmptyValue(value)) {
                            int col = getColumnIndexForField(fields[i]);
                            tableModel.setValueAt(value, row, col);
                            addRowToCache(row, col);
                        }
                    }
                }
            }
            setListeningForChanges(true);
        }
    }  
    
    private boolean isEmptyValue(Object o) {
        return (   o == null
        		|| StringUtils.isEmpty(o.toString())
				|| o.toString().equals("-1")
				|| o.toString().equals("0"));
    }

    //*************************************************************************
    //Private methods and classes
    //*************************************************************************
    private void buildTable() {
        DcObject dco = module.getDCObject();

        JTextField text = ComponentFactory.getTextField();
        
        Collection fields = dco.getFields();
        tableModel.setColumnCount(fields.size());
        
        int counter = 0;
        for (Iterator iter = fields.iterator(); iter.hasNext(); counter++) {
            DcField field = (DcField) iter.next();
            TableColumn columnNew = getColumnModel().getColumn(counter);
            columnNew.setIdentifier(new Integer(field.getIndex()));
            columnNew.setHeaderValue(field.getLabel());

            JComponent comp = ComponentFactory.getComponent(field.getFieldType(),
                                                            field.getLabel(),
                                                            field.getMaximumLength());
            comp.setAutoscrolls(false);
            comp.setBorder(null);
            comp.setDoubleBuffered(true);
            comp.setIgnoreRepaint(true);
            comp.setVerifyInputWhenFocusTarget(false);
            
            if (field.getFieldType() == ComponentFactory._LONGTEXTFIELD) {
                comp = ComponentFactory.getComponent(ComponentFactory._SHORTTEXTFIELD,
                                                     field.getLabel(),
                                                     field.getMaximumLength());
            }
            
            if (field.getIndex() == DcObject._ID) {
                text = ComponentFactory.getTextFieldDisabled();
                columnNew.setCellEditor(new DefaultCellEditor(text));
                DcTableCellRenderer renderer = new DcTableCellRenderer();
                renderer.setFont(ComponentFactory.getSystemFont());
                columnNew.setCellRenderer(renderer);
            } else if (field.getIndex() == GenericObject._SYS_MODULE) {
                text = ComponentFactory.getTextFieldDisabled();
                columnNew.setCellEditor(new DefaultCellEditor(text));
                columnNew.setCellRenderer(new ModuleTableCellRenderer());
            } else {
                switch (field.getFieldType()) {
                    case ComponentFactory._NUMBERFIELD:
                        columnNew.setCellEditor(new DefaultCellEditor((JTextField) comp));
                        columnNew.setMaxWidth(100);
                        columnNew.setCellRenderer(new NumberTableCellRenderer());
                        break;
                    case ComponentFactory._LONGTEXTFIELD :
                    case ComponentFactory._SHORTTEXTFIELD :
                        columnNew.setCellEditor(new DefaultCellEditor((JTextField) comp));
                        break;
                    case ComponentFactory._TIMEFIELD :
                        TimeFieldTableCellRenderer timeRenderer = new TimeFieldTableCellRenderer();
                        DcNumberField numberField = ComponentFactory.getNumberField();
                        columnNew.setCellEditor(new DefaultCellEditor(numberField));
                        columnNew.setCellRenderer(timeRenderer);
                        break;
                    case ComponentFactory._URLFIELD :
                        columnNew.setCellEditor(new DefaultCellEditor((JTextField) comp));
                        DcTableCellRenderer renderer = new DcTableCellRenderer();
                        renderer.setForeground(new Color(0, 0, 255));
                        columnNew.setCellRenderer(renderer);
                        break;
                    case ComponentFactory._PICTUREFIELD :
                        text = ComponentFactory.getTextField();
                        text.setEditable(false);
                        text.setFont(ComponentFactory.getUnreadableFont());
                        columnNew.setCellEditor(new DefaultCellEditor(text));
                        columnNew.setCellRenderer(
                                new PictureTableCellRenderer(module.getIndex()));
                        break;                        
                    case ComponentFactory._BOOKCATEGORYCOMBO:                        
                    case ComponentFactory._SOFTWARECATEGORYCOMBO:
                        columnNew.setCellRenderer(new ComboBoxTableCellRenderer());
                        columnNew.setCellEditor(new DefaultCellEditor((JComboBox) comp));
                        break;
                    case ComponentFactory._RATINGCOMBOBOX:
                        columnNew.setMinWidth(70);
                        columnNew.setCellRenderer(new RatingTableCellRenderer());
                        columnNew.setCellEditor(new DefaultCellEditor((DcRatingComboBox) comp));
                        break;
                    case ComponentFactory._MP3GENRECOMBO:
                    case ComponentFactory._YESNOCOMBO:
                        columnNew.setCellEditor(new DefaultCellEditor((JComboBox) comp));
                        break;
                }    
                comp.setEnabled(!readonly);
            }
        }

        createRowIndexColumn();
        addSystemColumns();
        
        setProperties();   
        
        applyDefinitions();
    }
    
    public void applyDefinitions() {
        Definitions definitions =
            DcSettings.getValueAsDefinitionGroup(module.getFieldDefinitionsSettingsKey());
        
        removeColumns();
        removeSystemColumns();
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            DcFieldDefinition definition = (DcFieldDefinition) iter.next();
            TableColumn column = (TableColumn) columns.get(new Integer(definition.getIndex()));  
            
            if (definition.isRequired()) 
                column.setHeaderRenderer(getRequiredTableHeaderRenderer());
            else
                column.setHeaderRenderer(getTableHeaderRenderer());
            
            String label = definition.getLabel();
            if (label != null && label.length() > 0) {
                column.setHeaderValue(label);
            } else {
                column.setHeaderValue(module.getField(definition.getIndex()).getSystemName());
            }
            
            if (definition.isVisible() && definition.isEnabled())
                addColumn(column);

        }
        addSystemColumns();
    }
    
    private void removeColumns() {
        for (Iterator iter = columns.values().iterator(); iter.hasNext(); ) {
            TableColumn column = (TableColumn) iter.next();
            try {
                removeColumn(column);
            } catch (Exception ignore) {}
        }        
    }
    
    /**
     * Adds all the columns to the table (hidden and visible fields)
     */
    public void resetTable() {
        
        Definitions definitions =
            DcSettings.getValueAsDefinitionGroup(module.getFieldDefinitionsSettingsKey());
        
        removeColumns();
        removeSystemColumns();
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            DcFieldDefinition definition = (DcFieldDefinition) iter.next();
            TableColumn column = (TableColumn) columns.get(new Integer(definition.getIndex()));
            addColumn(column);
        }
    }      
    
    /**
     * Applies all default properties on the table
     */
    private void setProperties() {
        setAutoscrolls(true);

        setDefaultRenderer(Object.class, new DcTableCellRenderer());
        setFont(ComponentFactory.getStandardFont());

        setRowSelectionAllowed(true);
        setColumnSelectionAllowed(false);
        setRequestFocusEnabled(true);
        setAutoResizeMode(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS);
        setAlignmentY(JTable.TOP_ALIGNMENT);
        getTableHeader().setReorderingAllowed(false);
        setShowHorizontalLines(true);
        
        setBackground(new Color(255, 255, 255));
        setGridColor(new Color(220,220,200));
        
        applyHeaders();

        setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
    }    
    
    /**
     * Enables / disables caching of changed values
     * @param b
     */
    private void setListeningForChanges(boolean b) {
        boolean enable = b && caching;
        if (enable && getListeners(tableChangeListener.getClass()).length == 0) {
            tableModel.addTableModelListener(tableChangeListener);
        } else if (!enable) {
            tableModel.removeTableModelListener(tableChangeListener);
        }
    }
    
    private DefaultTableCellRenderer getRequiredTableHeaderRenderer() {
        DcTableHeaderRendererRequired hdrRenderer = new DcTableHeaderRendererRequired();
        return hdrRenderer;
    }

    private DefaultTableCellRenderer getTableHeaderRenderer() {
        DcTableHeaderRenderer hdrRenderer = new DcTableHeaderRenderer();
        return hdrRenderer;
    }
    
    private void createRowIndexColumn() {
        rowIndexColumn = new TableColumn();
        rowIndexColumn.setHeaderValue("#");
        rowIndexColumn.setMaxWidth(40);
        rowIndexColumn.setPreferredWidth(40);
        rowIndexColumn.setWidth(40);
        JTextField text = ComponentFactory.getTextField();
        text.setEditable(false);
        text.setEnabled(false);
        text.setBackground(DcSettings.getValueAsColor(Repository.settings.stColor));
        rowIndexColumn.setCellEditor(new DefaultCellEditor(text));
        DcTableRowIndexCellRenderer renderer = new DcTableRowIndexCellRenderer();
        renderer.setFont(ComponentFactory.getSystemFont());
        rowIndexColumn.setCellRenderer(renderer);
        rowIndexColumn.setHeaderRenderer(getTableHeaderRenderer());
    }
    
    private void addSystemColumns() {
        getColumnModel().addColumn(rowIndexColumn);
        moveColumn(getColumnCount() - 1, 0);
    }

    private void removeSystemColumns() {
    	getColumnModel().removeColumn(rowIndexColumn);
    }
    
    /**
     * Handles changes made to the table (on screen)
     */
    private class TableValueChangedAction implements TableModelListener {
        public void tableChanged(TableModelEvent e) {
            if (    !ignoreEdit &&
            		(e.getType() == TableModelEvent.INSERT || 
                     e.getType() == TableModelEvent.UPDATE)) {

                int row = getSelectedRow();
                addRowToCache(row, e.getColumn());
            }
        }
    }    
}
