/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.dialogs;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcDialog;
import net.sf.dc.core.Repository;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;

/**
 * This dialog allows the user to apply a value on a field 
 * to all items on screen
 * 
 * @author Robert Jan van der Waals
 * @since 1.3
 * @version 0.9
 */
public class UpdateAllDialog extends DcDialog {
    
    private JButton buttonApply;
    private JButton buttonCancel;
    
    private Map labels = new HashMap();
    private Map fields = new HashMap();
    
    private JCheckBox checkSelectedItemsOnly;
    
    private DcObject dco;
    private boolean success = false;
   
    /**
     * Creates and shows the update all dialog
     * @param o Data Crow Object used as a reference
     * @param parent the parent frame
     */
    public UpdateAllDialog(DcObject dco, JFrame parent) {
        super(parent);

        setHelpIndex("dc.Tools.UpdateAll");
        
        setTitle(DcResources.getText(Repository.labels.lblUpdateAll));
        this.dco = dco;
        buildDialog(dco);
        
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                success = false;
                done();
            }
        });

        setSize(DcSettings.getValueAsDimension(
                Repository.settings.stUpdateAllDialogSize));

        checkSelectedItemsOnly.setSelected(
                DcSettings.getValueAsBoolean(Repository.settings.stUpdateAllSelectedItemsOnly));
        
        setCenteredLocation();

        Color color = DcSettings.getValueAsColor(Repository.settings.stColor);
        setBackground(color);
        
        setModal(true);        
        setVisible(true);
    }
    
    public boolean isUpdateSelectedItemsOnly() {
    	return checkSelectedItemsOnly.isSelected();
    }
    
    /**
     * Returns if the update all action was successfull
     */
    public boolean isSuccess() {
        return success;   
    }

    /**
     * Returns the values. These values should be used to update 
     * all the items 
     */
    public DcObject getValues() {
        DcObject object = dco.getModule().getDCObject();
        int[] indices = dco.getFieldIndices();
        for (int i = 0; i < indices.length; i++) {
            DcField field = dco.getField(indices[i]);
            int fieldType = field.getFieldType() == ComponentFactory._LONGTEXTFIELD ? 
                                ComponentFactory._SHORTTEXTFIELD : field.getFieldType();
            
            JComponent component = (JComponent) fields.get(field);
            Object value = ComponentFactory.getComponentValue(fieldType, component);
            
            if (value != null && !value.toString().equals("")) 
                object.setValue(field.getIndex(), value);
        }
        
        return object;
    }  
    
    /**
     * Closes this dialog
     */
    private void done() {
        DcSettings.setValue(Repository.settings.stUpdateAllDialogSize, getSize());
        DcSettings.setValue(Repository.settings.stUpdateAllSelectedItemsOnly, 
                            new Boolean(checkSelectedItemsOnly.isSelected()));        
        dispose();
    }
    
    /**
     * Hidden fields are not displayed
     * @param indices
     */
    private void hideFields(Collection removeFields) {
        DcField field = dco.getField(DcObject._ID);
        JLabel label = (JLabel) labels.get(field);
        JComponent component = (JComponent) fields.get(field);
        
        label.setVisible(false);
        component.setVisible(false);
        
        for (Iterator iter = removeFields.iterator(); iter.hasNext(); ) {
            field = (DcField) iter.next();
            label = (JLabel) labels.get(field);
            component = (JComponent) fields.get(field);
            label.setVisible(false);
            component.setVisible(false);
        }
    }     
    
    private void buildDialog(DcObject o) {
        Color color = DcSettings.getValueAsColor(Repository.settings.stColor);
        
        //**********************************************************
        //Input panel
        //**********************************************************
        JPanel panelInput = new JPanel(false);
        panelInput.setLayout(Layout.getGBL());
        panelInput.setBackground(color);
        
        
        int[] indices = dco.getFieldIndices();
        for (int i = 0; i < indices.length; i++) {
            DcField field = dco.getField(indices[i]);
            JLabel label = ComponentFactory.getLabel(field.getLabel());

            int fieldType = field.getFieldType() == ComponentFactory._LONGTEXTFIELD ? 
                                ComponentFactory._SHORTTEXTFIELD : field.getFieldType();

            JComponent component = ComponentFactory.getComponent(fieldType, field.getLabel(), field.getMaximumLength());
            
            fields.put(field, component);
            labels.put(field, label);

            component.setPreferredSize(new Dimension(100, 20));
            
            panelInput.add(label       , Layout.getGBC(0, i, 1, 1, 1.0, 1.0
                    , GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                      new Insets(0, 5, 0, 5), 0, 0));            
            panelInput.add(component   , Layout.getGBC(1, i, 1, 1, 10.0, 10.0
                    , GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                      new Insets(0, 5, 0, 5), 0, 0));
        }
        
        //**********************************************************
        //Options panel
        //**********************************************************
        checkSelectedItemsOnly = ComponentFactory.getCheckBox(DcResources.getText(Repository.labels.lblSelectedItemsOnly)); 
        JPanel panelOptions = new JPanel(false);
        panelOptions.setLayout(Layout.getGBL());
        panelOptions.setBackground(color);
        
        panelOptions.add(checkSelectedItemsOnly    , Layout.getGBC(0, 0, 1, 1, 0.0, 0.0
                       , GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                         new Insets(5, 5, 5, 5), 0, 0));            
        
        //**********************************************************
        //Action panel
        //**********************************************************
        JPanel panelActions = new JPanel(false);
        panelActions.setLayout(Layout.getGBL());
        panelActions.setBackground(color);

        buttonApply = ComponentFactory.getButton(DcResources.getText(Repository.labels.lblApply));
        buttonCancel = ComponentFactory.getButton(DcResources.getText(Repository.labels.lblCancel));
        
        buttonApply.addActionListener(new ApplyAction());
        buttonCancel.addActionListener(new CancelAction());
        
        buttonApply.setMnemonic(KeyEvent.VK_A);
        buttonCancel.setMnemonic(KeyEvent.VK_C);
        
        panelActions.add(buttonApply    , Layout.getGBC(0, 0, 1, 1, 0.0, 0.0
                                        , GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                                          new Insets(5, 5, 5, 5), 0, 0));            
        panelActions.add(buttonCancel   , Layout.getGBC(1, 0, 1, 1, 0.0, 0.0
                                        , GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                                          new Insets(5, 5, 5, 5), 0, 0));
                                          
        //**********************************************************
        //Main panel
        //**********************************************************
        this.getContentPane().setLayout(Layout.getGBL());
        this.getContentPane().add(panelInput  ,Layout.getGBC(0, 0, 1, 1, 10.0, 10.0
                                              ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                               new Insets(5, 5, 5, 5), 0, 0));
        this.getContentPane().add(panelOptions,Layout.getGBC(0, 1, 1, 1, 1.0, 1.0
                                              ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                                               new Insets(10, 5, 10, 5), 0, 0));
        this.getContentPane().add(panelActions,Layout.getGBC(0, 2, 1, 1, 1.0, 1.0
                                              ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                                               new Insets(5, 5, 5, 5), 0, 0));
        
        this.getContentPane().setBackground(color);
        pack();
        
        Collection removeFields = new ArrayList();
        removeFields.addAll(o.getHiddenFields());
        removeFields.addAll(o.getNotSearchableFields());
        
        hideFields(removeFields);
    }
    
    private class CancelAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            success = false;
            done();
        }
    }
    
    private class ApplyAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            success = true;
            done();
        }
    }   
}
