/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.forms;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.ScrollPaneConstants;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.MainFrame;
import net.sf.dc.console.book.AmazonBookSearchForm;
import net.sf.dc.console.components.DcFrame;
import net.sf.dc.console.components.DcLongTextField;
import net.sf.dc.console.components.DcPictureField;
import net.sf.dc.console.components.panels.ChildListPanel;
import net.sf.dc.console.dialogs.ItemTypeDialog;
import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.console.dialogs.QuestionBox;
import net.sf.dc.console.movie.ImdbMovieSearchForm;
import net.sf.dc.console.software.AmazonSoftwareSearchForm;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.exceptions.ValidationException;
import net.sf.dc.helpsystem.JavaHelp;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.settings.definitions.DcFieldDefinition;
import net.sf.dc.settings.definitions.Definitions;
import net.sf.dc.wf.requests.CloseItemFormRequest;
import net.sf.dc.wf.requests.RemoveRowsRequest;
import net.sf.dc.wf.requests.UpdateRowRequest;
import net.sf.ext.util.StringUtils;


/**
 * Form used to represent one item in the database or to create
 * a new item. The item form is very generic as it is build up by 
 * using the information present in the module. 
 * Child items are displayed using the DCChildList component 
 * 
 * @see net.sf.dc.console.components.panels.ChildListPanel 
 * @author Robert Jan van der Waals
 * @since 1.4
 * @version 1.9
 */
public class ItemForm extends DcFrame {
    
    private DcObject dco;
    private int moduleIndex;

    private Map labels = new HashMap();
    private Map fields = new HashMap();

    private JTabbedPane tabbedPane = new JTabbedPane();
    
    private boolean hideFields = false;
    private final boolean update;
    private final boolean readonly;
    
    private ChildListPanel childList;

    /**
     * Initializes and shows the item form 
     * 
     * @param readonly specifies if data can be edited / changed (and saved)
     * @param updateMode Update / Insert mode
     * @param dco Data Crow Object to be displayed
     * @param imdb use imdb update facilities
     * @param amazon use amazon update facilitie
     * @param clazz the clazz which triggered this item to be opened
     * @param row the UI row number
     */
    public ItemForm(    JFrame parent,
                        boolean readonly,
                        boolean update,
                        DcObject o, 
                        boolean bImdb,
                        boolean bAmazon) {
    	
        super(parent);

        Color color = DcSettings.getValueAsColor(Repository.settings.stColor);
        JavaHelp.setHelpIndex("dc.ItemForm");
        
        this.readonly = readonly;
        this.update = update;
        
        if (!update && !readonly &&  o.getModule().getIndex() == DCModules._GENERIC) {
            ItemTypeDialog dialog = new ItemTypeDialog(this);
            dialog.setVisible(true);
            this.moduleIndex = dialog.getSelectedModule();
            this.dco = DCModules.getModule(moduleIndex).getDCObject();
        } else {
            this.dco = o;
            this.moduleIndex = dco.getModule().getIndex();
        }
        
        this.hideFields = !DcSettings.getValueAsBoolean(Repository.settings.stShowHiddenFieldsInForm);
        
        setResizable(true);

        setIconImage(IconLibrary._icoMain.getImage());
        setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        
        if (!update) {
            setTitle(DcResources.getText(Repository.labels.lblNewItem, dco.getModule().getObjectName()));
        } else {
            setTitle(DcResources.getText(Repository.labels.lblItem, 
                     new String[] {dco.getModule().getObjectName(), dco.getDisplayString(1)}));
        }
        
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                close(false);
            }
        });
        
        this.getContentPane().setBackground(color);
        this.setBackground(color);
        this.getContentPane().setLayout(Layout.getGBL());
        
        initializeComponents();

        DCModule module = DCModules.getModule(moduleIndex);
        JPanel panelActions = getActionPanel(readonly, module.isAmazonUpdateAllowed(), module.isImdbUpdateAllowed());
        
        addInputPanel();
        addChildrenPanel();        
        addPictureTabs(); 
        
        getContentPane().add(tabbedPane,  Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                             new Insets(0,0,15,0), 0, 0));        
        getContentPane().add(panelActions,  Layout.getGBC(0, 1, 1, 1, 0.0, 0.0
                            ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                             new Insets(5, 5, 5, 5), 0, 0));        

        setRequiredFields();
        setReadonly(readonly);
        setData(dco, true);        
        
        pack();
        
        setSize(DcSettings.getValueAsDimension(getItemSizeSetting()));
            
        setCenteredLocation();
    }
    
    public JTabbedPane getTabbedPane() {
        return tabbedPane;
    }
    
    /**
     * Closes this form by setting the visibility to false
     */
    public void close(boolean aftersave) {
        if (!aftersave && update && dco != null && (!readonly && isChanged())) {
        	new QuestionBox(DcResources.getText(Repository.messages.msgNotSaved));
        	if (QuestionBox.result != null && QuestionBox.result.booleanValue()) {
        		saveValues();
        	}
        }
    	
    	DcSettings.setValue(getItemSizeSetting(), getSize());
        
        fields = null;
        labels = null;
        childList = null;
        tabbedPane = null;
        dispose();
    }  
    
    /**
     * Applies data to this form
     * @param values    values to be displayed. Values is based on all fields, both
     *                  hidden and visible
     * @param overwrite indicates if existing values should be overwritten
     * @param destinationFields the field indices which are targeted to be filled
     *                  with the values
     */
    public void setData(DcObject object, boolean overwrite) {
        if (childList != null) {
            dco.setChildren(object.getChildren());
            childList.setChildren(object.getChildren());
        }
        
        int[] indices = object.getFieldIndices();
        for (int i = 0; i < indices.length; i++) {
            int index = indices[i];
            DcField field = dco.getField(index);
            JComponent component = (JComponent) fields.get(field);
            
            Object oldValue = ComponentFactory.getComponentValue(field.getFieldType(), component);
            Object newValue = object.getValue(index);
            
            boolean empty = oldValue == null || StringUtils.isEmpty(oldValue.toString()) && !oldValue.equals("-1");
            
            if ((empty || overwrite) && !newValue.equals("")) {
                ComponentFactory.setComponentValue(field.getFieldType(), component, newValue);
            }
        }
    }    
    
    /**
     * Sets all the fields in this form to readonly
     * @param readonly
     */
    private void setReadonly(boolean readonly) {
        if (readonly) {
            for (Iterator iter = fields.values().iterator(); iter.hasNext(); ) {
                JComponent component = (JComponent) iter.next();
                component.setEnabled(false);
                component.setBackground(ComponentFactory.getDisabledColor());
            }
        }        
    }
    
    /**
     * Marks fields as being required. The label is set to the required field 
     * color
     */
    private void setRequiredFields() {
        int[] fields = dco.getFieldIndices();
        for (int i = 0; i < fields.length; i++) {
            DcField field = dco.getField(fields[i]);
            if (field.isRequired()) {
                JLabel label = (JLabel) labels.get(field);
                label.setForeground(ComponentFactory.getRequiredColor());
            }
        }  
    }    
    
    /**
     * Retrieves the size settings key for the current active module
     */
    private String getItemSizeSetting() {
        String setting = Repository.settings.stItemFormViewSize;
        if (moduleIndex == DCModules._MOVIE) {
            setting = Repository.settings.stMovieItemFormViewSize;
        } else if (moduleIndex == DCModules._SOFTWARE) {
            setting = Repository.settings.stSoftwareItemFormViewSize;
        } else if (moduleIndex == DCModules._SOFTWARECAT) {
            setting = Repository.settings.stSoftwareCatItemFormViewSize;
        } else if (moduleIndex == DCModules._MUSIC) {
            setting = Repository.settings.stMusicItemFormViewSize;
        } 
        return setting;
    }

    /**
     * Starts a Amazon update. The update will show its own dialog
     * and transfers data selected by the user back to this item
     * form.
     */
    private void startAmazonUpdate() {
        AmazonSearchForm amazonForm = null;
        if (moduleIndex == DCModules._SOFTWARE) {
            amazonForm = new AmazonSoftwareSearchForm(dco, this);
        } else {
            amazonForm = new AmazonBookSearchForm(dco, this);            
        }
        amazonForm.setVisible(true);
        
        amazonForm.setVisible(true);
        amazonForm.startSearch();
    }

    /**
     * Starts a IMDB update. The update will show its own dialog
     * and transfers data selected by the user back to this item
     * form.
     */
    private void startIMDBUpdate() {
        ImdbMovieSearchForm imdbForm = new ImdbMovieSearchForm(dco, this);
        imdbForm.setVisible(true);
        imdbForm.startSearch();
    }

    /**
     * Deletes this item from the database
     */
    private void deleteValues() {
        new QuestionBox(DcResources.getText(Repository.messages.msgDeleteQuestion), this);
        
        if (QuestionBox.result.booleanValue()) {
            String id = dco.getID();
            dco.clearValues();
            dco.setValue(DcObject._ID, id);
            dco.setSilence(true);

            int tab = update ? MainFrame._SEARCHTAB : MainFrame._INSERTTAB;
            dco.addRequest(new CloseItemFormRequest(this));
            dco.addRequest(new RemoveRowsRequest(moduleIndex, tab, new String[] {id}));
            dco.delete();
        }
    }
    
    private boolean isChanged() {
        boolean changed = dco.isChanged();
        
        int[] indices = dco.getFieldIndices();
        for (int i = 0; i < indices.length; i++) {
            int index = indices[i];
            
            DcField field = dco.getField(index);
            JComponent component = (JComponent) fields.get(field);
            
            Object o = ComponentFactory.getComponentValue(field.getFieldType(), component);
            
            String newValue = o == null ? "" : o.toString();
            newValue = newValue.equals("-1") || newValue.equals("0") ? "" : newValue;

            String oldValue = dco.getValue(index).toString();
            oldValue = oldValue.equals("-1") || oldValue.equals("0") ? "" : oldValue;
            
            if (!oldValue.toString().equals(newValue)) {
                changed = true;
                break;
            }
        }
        return changed;
    }    
    
    public DcObject getObject() {
        dco.removeChildren();
        if (childList != null) {
            DcObject[] children = childList.getChildren();
            for (int i = 0; i < children.length; i++) {
                dco.addChild(children[i]);
            }
        }

        if (update) dco.markAsUnchanged();
        
        for (Iterator iter = fields.keySet().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            JComponent component = (JComponent) fields.get(field);
            
            Object value = ComponentFactory.getComponentValue(field.getFieldType(), component);
            value = value == null ? "" : value;
            if (update) {
                Object oldValue = dco.getValue(field.getIndex());
                if (!oldValue.equals(value)) 
                    dco.setValue(field.getIndex(), value);
                
            } else if (isChanged()) {
                if (!value.equals("")) 
                    dco.setValue(field.getIndex(), value);
                
                dco.setValue(DcObject._ID, "");
            }
        }
        dco.setSilence(true);
        return dco;
    }

    /**
     * Inserts or updates the Data Crow object
     */
    private void saveValues() {
        dco = getObject();
        
        if (!update)  {
            try {
                dco.addRequest(new CloseItemFormRequest(this));
                dco.setSilence(true);
                dco.saveNew();
            } catch (ValidationException vExp) {
                // when an error is encountered, saving is stopped
                new MessageBox(vExp.getMessage(), MessageBox._WARNING);
                return;
            }
        } else if (isChanged()) {
            int tab = update ? MainFrame._SEARCHTAB : MainFrame._INSERTTAB;
            dco.addRequest(new UpdateRowRequest(tab, dco, true, true, false));
            try {
                dco.addRequest(new CloseItemFormRequest(this));
                dco.setSilence(true);
                dco.saveUpdate();
            } catch (ValidationException vExp) {
                new MessageBox(vExp.getMessage(), MessageBox._WARNING);
                return;
            }
        } else if (! isChanged()) {
        	new MessageBox(DcResources.getText(Repository.messages.msgNoChangesToSave), MessageBox._WARNING);
        }
    }
    
    private void initializeComponents() {
        int[] indices = dco.getFieldIndices();
        for (int i = 0; i < indices.length; i ++) {
            int index = indices[i];
            DcField field = dco.getField(index);

            labels.put(field, ComponentFactory.getLabel(dco.getLabel(index)));
            if (index == DcObject._ID) {
                fields.put(field, ComponentFactory.getTextFieldDisabled());
            } else {
                JComponent c = ComponentFactory.getComponent(field.getFieldType(),  
                                                             field.getLabel(),
                                                             field.getMaximumLength());
                fields.put(field, c);
            }
        }
    }
    
    private void addInputPanel() {
        Color color = DcSettings.getValueAsColor(Repository.settings.stColor);
        
        JPanel panelInfo = new JPanel(); 
        JPanel panelTech = new JPanel();
        panelInfo.setLayout(Layout.getGBL());
        panelInfo.setBackground(color);
        panelTech.setLayout(Layout.getGBL());
        panelTech.setBackground(color);
        
        DCModule module = DCModules.getModule(moduleIndex);
        Definitions definitions = DcSettings.getValueAsDefinitionGroup(module.getFieldDefinitionsSettingsKey());
      
        int yTech = 0;
        int yInfo = 0;
        boolean technicalInfoPresent = false;
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            DcFieldDefinition definition = (DcFieldDefinition) iter.next();
            int index = definition.getIndex();
            
            DcField field = dco.getField(index);
            JLabel label = (JLabel) labels.get(field);
            JComponent component = (JComponent) fields.get(field);
        	if ( definition.isEnabled() && // enabled
                ((hideFields && definition.isVisible()) || !hideFields) && // check this visibility
                !(component instanceof DcPictureField) && // check the field type
                index != dco.getParentReferenceFieldIndex()) { // not a reference field
                
                int stretch = GridBagConstraints.HORIZONTAL;
                int factor = 1;
                
                if (field.getFieldType() == ComponentFactory._LONGTEXTFIELD) {
                    stretch = GridBagConstraints.BOTH; 
                    factor  = 50;
    
                    DcLongTextField longText = (DcLongTextField) component;
                    longText.setMargin(new Insets(1, 1, 1, 5));

                    JScrollPane pane = new JScrollPane(longText);
                    pane.setBorder(ComponentFactory.getBorder());
                    pane.setPreferredSize(new Dimension(100,100));                    
                    pane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
                    pane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
                    component = pane;
                } else {
                	component.setPreferredSize(new Dimension(200, 20));
                }
                
                label.setPreferredSize(new Dimension(100, 20));
                label.setBackground(color);
                
                if (!field.isTechnicalInfo()) {
                    panelInfo.add(label, Layout.getGBC(0, yInfo, 1, 1, 1.0, 1.0
                             ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                              new Insets(0, 0, 0, 5), 0, 0));
                    panelInfo.add(component,  Layout.getGBC(1, yInfo, 1, 1, factor, factor
                             ,GridBagConstraints.NORTHWEST, stretch,
                              new Insets(0, 0, 0, 0), 0, 0));
                    yInfo++;
                } else {
                    technicalInfoPresent = true;
                    panelTech.add(label, Layout.getGBC(0, yTech, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                             new Insets(0, 0, 0, 5), 0, 0));
                    panelTech.add(component,  Layout.getGBC(1, yTech, 1, 1, factor, factor
                            ,GridBagConstraints.NORTHWEST, stretch,
                             new Insets(0, 0, 0, 0), 0, 0));
                    yTech++;
                }
                
            }
        }
        tabbedPane.addTab(DcResources.getText(Repository.labels.lblDescription), IconLibrary._icoDescription, panelInfo);
        if (technicalInfoPresent) {
            JPanel dummy = new JPanel();
            dummy.setLayout(Layout.getGBL());
            dummy.setBackground(color);
            dummy.add(panelTech,  Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                      ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                      new Insets(2, 0, 0, 0), 0, 0));
            tabbedPane.addTab(DcResources.getText(Repository.labels.lblTechnicalInfo), IconLibrary._icoTechInfo, dummy);
        }
    }
    
    private void addChildrenPanel() {
        DCModule module = DCModules.getModule(moduleIndex);
        DCModule childModule = module.getChildModule();
        if (childModule != null) {
            childList = new ChildListPanel(childModule);
            tabbedPane.addTab(childModule.getObjectName() + "s", childList);
        }
    }
    
    private void addPictureTabs() {
        DCModule module = DCModules.getModule(moduleIndex);
        Color color = DcSettings.getValueAsColor(Repository.settings.stColor);
        
        Definitions definitions = DcSettings.getValueAsDefinitionGroup(module.getFieldDefinitionsSettingsKey());
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            DcFieldDefinition definition = (DcFieldDefinition) iter.next();

            int index = definition.getIndex();
            DcField field = dco.getField(index);
            JComponent component = (JComponent) fields.get(field);            
            
            if (definition.isEnabled() &&
                ((hideFields && definition.isVisible()) || !hideFields) && 
                (component instanceof DcPictureField)) {
                
                JPanel panel = new JPanel(); 
                panel.setLayout(Layout.getGBL());
                panel.setBackground(color);
                
                component.setPreferredSize(component.getMinimumSize());
                panel.add(component, Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                         ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                          new Insets(5, 5, 5, 5), 0, 0));
                
                tabbedPane.addTab(field.getLabel(), IconLibrary._icoPicture, panel);
            }
        }
    }
    
    private JPanel getActionPanel(boolean readonly, boolean amazon, boolean imdb) {
        
        DCModule module = DCModules.getModule(moduleIndex);
        Color color = DcSettings.getValueAsColor(Repository.settings.stColor);

        JPanel panel = new JPanel(false);
        panel.setBackground(color);
        panel.setLayout(Layout.getGBL());
        
        String name = module.getObjectName();
        JButton buttonClose = ComponentFactory.getButton(DcResources.getText(Repository.labels.lblClose));
        JButton buttonDelete = ComponentFactory.getButton(DcResources.getText(Repository.labels.lblDelete, name));
        JButton buttonInternet = null;
        
        if (imdb) {
            buttonInternet = ComponentFactory.getButton(DcResources.getText(Repository.labels.lblIMDBUpdate));
            buttonInternet.addActionListener(new IMDBAction());
            buttonInternet.setToolTipText(DcResources.getText(Repository.tooltips.tpIMDBUpdate));
            buttonInternet.setMnemonic(KeyEvent.VK_U);
        } else if (amazon) {
            buttonInternet = ComponentFactory.getButton(
                             DcResources.getText(Repository.labels.lblAmazonUpdate));
            buttonInternet.addActionListener(new AmazonAction());
            buttonInternet.setToolTipText(DcResources.getText(Repository.tooltips.tpAmazonUpdate));
            buttonInternet.setMnemonic(KeyEvent.VK_U);
        }        

        JButton buttonSave = ComponentFactory.getButton(DcResources.getText(Repository.labels.lblSave, name));
        buttonSave.setMnemonic(KeyEvent.VK_S);
        buttonSave.setToolTipText(DcResources.getText(Repository.tooltips.tpSave));
        buttonSave.addActionListener(new SaveAction());
        
        buttonClose.setMnemonic(KeyEvent.VK_C);
        buttonClose.setToolTipText(DcResources.getText(Repository.tooltips.tpClose));
        buttonClose.addActionListener(new ExitAction());
        
        buttonDelete.setMnemonic(KeyEvent.VK_D);
        buttonDelete.setToolTipText(DcResources.getText(Repository.tooltips.tpDelete));
        buttonDelete.addActionListener(new DeleteAction());
        
        if (!readonly) {
        panel.add(buttonSave, Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                 ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                  new Insets(5, 5, 5, 5), 0, 0));
        }
                        
        if (amazon || imdb) {
            panel.add(buttonInternet, Layout.getGBC(1, 0, 1, 1, 1.0, 1.0
                     ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                      new Insets(5, 5, 5, 5), 0, 0));
        }
        
        if (update && !readonly) {
            panel.add(buttonDelete, Layout.getGBC(2, 0, 1, 1, 1.0, 1.0
                     ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                      new Insets(5, 5, 5, 5), 0, 0));
        }
        
        panel.add(buttonClose , Layout.getGBC(3, 0, 1, 1, 1.0, 1.0
                 ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                  new Insets(5, 5, 5, 5), 0, 0));
        
        return panel;
    }

    private class IMDBAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            startIMDBUpdate();
        }
    }

    private class AmazonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            startAmazonUpdate();
        }
    }

    private class SaveAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            saveValues();
        }
    }

    private class DeleteAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            deleteValues();
        }
    }

    private class ExitAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            close(false);
        }
    }
}
