/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.modules;

import java.awt.Image;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.swing.ImageIcon;
import javax.swing.JMenu;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.components.DcMenuBar;
import net.sf.dc.console.components.DcPanel;
import net.sf.dc.console.components.panels.QuickViewPanel;
import net.sf.dc.console.components.panels.TableInsertPanel;
import net.sf.dc.console.components.panels.TableSearchPanel;
import net.sf.dc.console.forms.SearchForm;
import net.sf.dc.console.importcd.ImportCDDialog;
import net.sf.dc.core.Repository;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.processes.cdimport.ImportCDProcess;
import net.sf.dc.settings.definitions.ReportFieldDefinitions;

/**
 * Contains all generic information for this module. 
 * Only static data can be found in here, such as the column and
 * table information. The UI representation can also be retrieved
 * thru this class (insert and search panels)
 * 
 * Hidden field information and such, which can be changed during
 * a session is contained within the Data Crow Objects themselves.
 * 
 * Main reason for this setup is to keep the Data Crow Objects as
 * small as possible, they can be stored in memory by thousands
 * at a time..
 * 
 * @see net.sf.dc.core.modules.DCModules
 * @see net.sf.dc.core.modules.DCModule
 * 
 * @author Robert Jan van der Waals
 * @since 1.4
 * @version 2.4
 */
public abstract class DCModule {
    
    protected TableInsertPanel insertPanel;
    protected TableSearchPanel searchPanel;
    protected SearchForm searchForm;
    
    protected DCModule childModule;
    protected Map fields = new HashMap();
    private Map systemFields = new HashMap();
    
    private final int index;
    private final String name;
    protected QuickViewPanel quickView;
    
    private final DcMenuBar menuBar;
    
    private ImageIcon iconSmall = null;
    
    private final boolean topModule;
    
    private final boolean allowImdbUpdate;
    private final boolean allowAmazonUpdate;
    private final boolean allowFreeDbUpdate;
    
    protected String tableName = "";
    protected String tableShortName = "";
    protected String tableJoin = "";
    
    /**
     * Initializes this module
     * 
     * @param index static final module index
     * @param name name of this module
     * @param allowImdbUpdate internet updates, thru imdb, 
     *        allowed y/n
     * @param allowAmazonUpdate internet updates, thru amazon, 
     *        allowed y/n
     * @param standalone is this a standalone or a composite module 
     */
    public DCModule(        int index, 
                            boolean topModule,
                            String name,
                            String tableName,
                            String tableShortName,
                            String tableJoin,
                            boolean allowImdbUpdate,
                            boolean allowAmazonUpdate,
                            boolean allowFreeDbUpdate) {
        this.index = index;
        this.tableName = tableName;
        this.tableShortName  = tableShortName;
        this.tableJoin = tableJoin;
        this.name = name;
        this.topModule = topModule;
        this.allowAmazonUpdate = allowAmazonUpdate;
        this.allowImdbUpdate = allowImdbUpdate;
        this.allowFreeDbUpdate = allowFreeDbUpdate;

        initializeSystemFields();
        initializeFields();

        this.menuBar = new DcMenuBar(this);
    }
    
    private void initializeSystemFields() {
        systemFields.put(new Integer(DcObject._SYS_MODULE), 
                         new DcField(DcObject._SYS_MODULE, getIndex(), "Item", 
                                     true, true, true, false, false, 
                                     255, ComponentFactory._SHORTTEXTFIELD, Repository.datatypes._STRING, 
                                     "ItemType"));
    }

    
    /**
     * Returns all insert and search panels belonging to this module
     * @return array of panels
     */
    public DcPanel[] getPanels() {
        if (getSearchPanel() != null && getInsertPanel() != null) {
            return new DcPanel[] {getSearchPanel(), getInsertPanel()};
        } else if (getSearchPanel() != null) {
             return new DcPanel[] {getSearchPanel()};
        } else if (getInsertPanel() != null) {
            return new DcPanel[] {getInsertPanel()};
        } else {
        	return new DcPanel[0];
        }
    }
    
    public DcField getField(int index) {
        DcField field = (DcField) fields.get(new Integer(index));
        return field == null ? getSystemField(index) : field;
    }
    
    public DcField getSystemField(int index) {
        return (DcField) systemFields.get(new Integer(index));
    }    

    /**
     * Sets a child module for this module
     * @param childModule child module
     */
    public void setChildModule(DCModule childModule) {
    	this.childModule = childModule;
    }
    
    public DcMenuBar getMenuBar() {
        return menuBar;
    }    
    
    /**
     * Returns the child module (or null)
     */
    public DCModule getChildModule() {
    	return childModule;
    }
    
    public SearchForm getSearchForm() {
        if (searchForm == null) searchForm = new SearchForm(this, searchPanel);
        return searchForm;
    }

    public Collection getFields() {
        return fields.values();
    }

    public int[] getFieldIndices() {
        Set keys = fields.keySet();
        int[] indices = new int[keys.size()];
        int counter = 0;        
        for (Iterator iter = keys.iterator(); iter.hasNext(); counter++) {
            indices[counter] = ((Integer) iter.next()).intValue();
        }
        return indices;
    }
    
    protected void initializeFields() {
        fields.put( new Integer(DcObject._ID), 
                    new DcField(DcObject._ID, getIndex(), "ID", 
                                false, true, true, false, false,
                                15, ComponentFactory._SHORTTEXTFIELD, Repository.datatypes._STRING, 
                                "ID"));    
    }
    
    public boolean isTopModule() {
        return topModule;
    }
    
    public int getFieldCount() {
        return fields.size();
    }
    
    public ImageIcon getIconSmall() {
        if (iconSmall == null) { 
            iconSmall = getIcon();
            Image image = iconSmall.getImage();
            
            int width = 32, height = 32;
            if (image != null) {
                width = iconSmall.getIconWidth() / 2;
                height = iconSmall.getIconHeight() / 2;
            }
            iconSmall = new ImageIcon(image.getScaledInstance(width, height, Image.SCALE_DEFAULT));
        }
        
        return iconSmall;
    }
    
    public abstract ImageIcon getIcon();
    public abstract ReportFieldDefinitions getReportFieldDefinitions();
    public abstract String getFieldDefinitionsSettingsKey();
    public abstract String getObjectName();
    public abstract QuickViewPanel getQuickView();
    public abstract DcObject[] getDefaultData() throws Exception;
    public abstract void startImport();

    /**
     * Returns an empty Data Crow object, managed by this module 
     */
    public abstract DcObject getDCObject();
    
    /**
     * Returns module specific menu items  
     */
    public abstract JMenu getToolsMenu();
    
    /**
     * Creates the panels for this module
     */
    protected abstract void initializeUI();
    
    /**
     * Returns the module specific CD import process
     * 
     * @param sourceDir directory which is used as source for the
     *        import
     * @param confirmEachDir indicates whether for each directory
     *        confirmation should be asked if the underlying data
     *        of this directory should be read or not.
     * @param nameOfCD cd name (used as the location name)
     */
    public abstract ImportCDProcess getCDImportProcess(ImportCDDialog ui,
                                                       String sourceDir,
													   boolean confirmDirs,
                                                       boolean recursiveDirs,
													   String nameOfCD);

    /**
     * Initializes the insert panel (lazy)
     */
    public TableInsertPanel getInsertPanel() {
        if (insertPanel == null && searchPanel == null) 
            initializeUI();
    
        return insertPanel;
    }
    
    /**
     * Initializes the search panel (lazy)
     */
    public TableSearchPanel getSearchPanel() {
        if (insertPanel == null && searchPanel == null) 
            initializeUI();

        return searchPanel;
    }
    
    /**
     * Indicates if this module has an Amazon update implementation
     */
    public boolean isFreedBUpdateAllowed() {
        return allowFreeDbUpdate;
    }
    
    /**
     * Indicates if this module has an Amazon update implementation
     */
    public boolean isAmazonUpdateAllowed() {
        return allowAmazonUpdate;
    }

    /**
     * Indicates if this module has an IMDB update implementation
     */
    public boolean isImdbUpdateAllowed() {
        return allowImdbUpdate;
    }
    
    /**
     * Returns the display name of this module
     */
    public String getName() {
        return name;
    }
    
    public String getTableName() {
    	return tableName;
    }

    public String getTableShortName() {
        return tableShortName;
    }
    
    public String getTableJoin() {
    	return tableJoin;
    }
    
    /**
     * Returns the index for this module as defined in DCModules
     * @see DCModules 
     */
    public int getIndex() {
        return index;
    }
    
    /**
     * Creates a string representation for this module
     */
    public String toString() {
        return name;
    }
}
