/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.objects;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.swing.ImageIcon;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.db.QueryOptions;
import net.sf.dc.exceptions.ValidationException;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.settings.definitions.DcFieldDefinition;
import net.sf.dc.settings.definitions.DcFieldDefinitions;
import net.sf.dc.wf.WorkFlow;
import net.sf.dc.wf.requests.IRequest;
import net.sf.dc.wf.requests.RequestCollection;
import net.sf.ext.util.Utilities;

/**
 * Base class for all Data Crow objects. Here the fields, values and required
 * fields are stored. Checks are done to garantee uniqueness and validity
 * of the items.
 *
 * Objects which extend this class need to override the save and load methods
 *
 * @author Robert Jan van der Waals
 * @since 1.3
 * @version 0.5
 */
public abstract class DcObject {

    protected int uiRowNumber = -1;
    protected int module;

    protected Map values = new HashMap();
    
    /** Indicates of the data should be checked for validity */
    protected boolean bValidation = true;
    
    /** Indicates if queries should display output */
    private boolean bSilence = false;
    private RequestCollection requests = new RequestCollection();    
    protected static boolean alwaysGenerateID = false;
    
    protected Collection alChildren = new ArrayList();
    
    private boolean isBatch = false; 
    private boolean endOfBatch = false;
    
    public static final int _ID = 0;
    public static final int _SYS_MODULE = 201;  
    
    /**
     * Creates an empty data crow object
     */
    protected DcObject() {
        setModuleIndex();
        initializeValues();
        initializeSystemValues();
        String definitionsKey = DCModules.getModule(module).getFieldDefinitionsSettingsKey();
        applySettings((DcFieldDefinitions) DcSettings.getValue(definitionsKey));
    }

    /**
     * Remove all requestors listening in on this object
     */
    public void removeRequestors() {
        requests.clear();
    }
    
    public void setValueForColumn(String column, Object value) {
        Collection fields = getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.getDatabaseFieldName().equalsIgnoreCase(column)) {
                setValue(field.getIndex(), value);
                break;
            }
        }
    }
    
    /**
     * Returns all requestors
     */
    public RequestCollection getRequests() {
    	return requests;
    }
    
    public String getTableName() {
    	return getModule().getTableName();
    }

    public String getTableShortName() {
        return getModule().getTableShortName();
    }
    
    public String getTableJoin() {
        return getModule().getTableJoin();
    }    
    
    public String getLabel(int index) {
        return getField(index).getLabel();
    }
    
    /**
     * Add a requestor. The requestor will be executed once the object
     * is saved / deleted or updated
     * @param requestor
     */
    public void addRequest(IRequest request) {
        requests.addRequest(request);
    }
    
    /**
     * Indicates this object to be the last in row of an update / insert or delete
     * @param end
     */
    public void setEndOfBatch(boolean end) {
    	endOfBatch = end;
    }
    
    /**
     * Indicates this object being part of a larger operation
     * @param batch part of batch y/n
     */
    public void setBatch(boolean batch) {
    	isBatch = batch;
    }
    
    /**
     * Part of a batch ?
     */
    public boolean isBatch() {
        return isBatch;
    }

    /**
     * End of the batch ?
     */
    public boolean isEndOfBatch() {
    	return endOfBatch;
    }
    
    /**
     * Remove all children from this object
     */
    public void removeChildren() {
    	alChildren.clear();
        alChildren = new ArrayList();
    }

    public void setChildren(Collection children) {
        alChildren = children;
    }

    
    public void addChildren(Collection children) {
        alChildren.addAll(children);
    }
    
    /**
     * Adds a child to this object
     * @param child
     */
    public void addChild(DcObject child) {
        child.setBatch(isBatch());
        
        if (endOfBatch) {
        	setEndOfBatch(false);
            
            for (Iterator iter = alChildren.iterator(); iter.hasNext(); ) {
                ((DcObject) iter.next()).setEndOfBatch(false);    
            }
            child.setEndOfBatch(true);
        }
        
        if (!alChildren.contains(child)) {
            alChildren.add(child);
        }
    }
    
    /**
     * Retrieve all children of this object
     */
    public Collection getChildren() {
        return alChildren;
    }

    /**
     * Retrieve the field index holding the reference to the parent 
     * (the primary key of the parent)
     */
    public int getParentReferenceFieldIndex() {
        Collection fields = getModule().getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
        	if (field.getValueType() == Repository.datatypes._PARENTREFERENCE) {
        		return field.getIndex();
            }
        }
        return -1;
    }
    
    /**
     * Returns the icon for this object. Example; Software Categories
     * has an icon  
     */
    public abstract ImageIcon getIcon();
    
    /**
     * Returns the filename
     */
    public abstract String getFilename();
    
    /**
     * Returns the module to which this object belongs.
     * Shortcut for getting the module object instead of the 
     * module index.
     */
    public DCModule getModule() {
        return DCModules.getModule(module);
    }
    
    /**
     * Returns all the field definitions. 
     * Shortcut; normally you would first have to ask for the
     * module and then ask to the module which field definitions
     * it contains. 
     */
    public Collection getFields() {
        return getModule().getFields();
    }
    
    /**
     * Sets all fields to 'unchanged'. This garantees that none of the
     * values will be saved to the database when performing an update
     * or insert. All 'setValue' calls are up to this moment are 
     * discarded
     */
    public void markAsUnchanged() {
        for (Iterator iter = values.values().iterator(); iter.hasNext(); ) {
            DcValue value = (DcValue) iter.next();
            value.setChanged(false);
        }
    }
    
    public void reload() {
        String query = "SELECT * FROM " + getTableName() + " WHERE ID = '" + getID() + "';";
        Collection objects = WorkFlow.runQueryDirect(query);
        for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
            DcObject dco = (DcObject) iter.next();
            int[] fields = getModule().getFieldIndices();
            for (int i = 0; i < fields.length; i++) {
                setValue(fields[i], dco.getValue(fields[i]));
            }
        }
        markAsUnchanged();
    }
    
    private void markAsChanged() {
        for (Iterator iter = values.values().iterator(); iter.hasNext(); ) {
            DcValue value = (DcValue) iter.next();
            value.setChanged(true);
        }
        
        for (Iterator iter = alChildren.iterator(); iter.hasNext(); ) {
        	DcObject child =  (DcObject) iter.next();
        	child.markAsChanged();
        }
    }    
    
    /**
     * Indicates if changes have been made to this data crow object.
     * A field is indicated as beign changed when the user or system
     * has performed a set value.
     */
    public boolean isChanged() {
        for (Iterator iter = values.values().iterator(); iter.hasNext(); ) {
            DcValue value = (DcValue) iter.next();
            if (value.isChanged()) {
                return true;
            }
        }        
        return false;
    }

    /**
     * Indicates if changes have been made to the specified value.
     * A field is indicated as beign changed when the user or system
     * has performed a set value.
     * 
     * @param index the value index to check on
     */
    public boolean isChanged(int index) {
        return getValueDef(index).isChanged();
    }
    
    /**
     * Marks the specified value as changed (or not) This means that 
     * the value will be considered for an insert or update to the 
     * database
     * 
     * @param index the value index to mark as changed / unchanged
     * @param b changed y/n 
     */    
    public void setChanged(int index, boolean b) {
        Object o = getValue(index);
        if (!o.equals(new Integer(-1))) {
        	getValueDef(index).setChanged(b);
        }
    }    
    
    /**
     * Indicates if the given field is visible to the user or not
     * @param index field index
     */
    public boolean isEnabled(int index) {
        return getModule().getField(index).isEnabled();
    }    

    /**
     * Indicates whether the given field is required or not
     * @param index field index
     */
    public boolean isRequired(int index) {
        return getModule().getField(index).isRequired();
    }

    /**
     * Indicates whether the given field is required or not
     * @param index field index
     */
    public boolean isSearchable(int index) {
        return getModule().getField(index).isSearchable();
    }    
    
    /**
     * Retrieves all disabled fields
     */
    public int[] getDisabledFields() {
        Collection fields = getModule().getFields();

        int counter = 0;
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (!field.isEnabled()) {
                counter++;
            }
        }

        
        int[] disabled = new int[counter];
        counter = 0;
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (!field.isEnabled()) {
                disabled[counter++] = field.getIndex();
            }
        }
        
        return disabled;
    }    
    
    
    /**
     * Retrieves all currently, to the user, hidden fields
     */
    public Collection getHiddenFields() {
        Collection fields = getModule().getFields();
        Collection hidden = new ArrayList();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.isHidden()) 
                hidden.add(field);
        }
        return hidden;
    }    

    /**
     * Retrieves the ID for this data crow object
     */
    public String getID() {
        return getValueDef(_ID).getValueAsString();
    }
    
    /**
     * Retrieves all fields on which a search cannot be performed.
     * Example; a user can not perform a search on a picture field
     */
    public Collection getNotSearchableFields() {
        Collection fields = getModule().getFields();
        Collection notSearchable = new ArrayList();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (!field.isSearchable()) 
                notSearchable.add(field);
        }
        return notSearchable;  
    }
    
    /**
     * Retrieves the indices of the, for the user visible fields
     */
    public Collection getVisibleFields() {
        Collection fields = getModule().getFields();
        Collection visible = new ArrayList();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (!field.isHidden()) 
                visible.add(field);
        }
        return visible;  
    }

    /**
     * Sets a value for the given field
     * @param iField field indicator
     * @param sValue value as String
     */
    public void setValue(int index, Object o) {
    	DcValue value = getValueDef(index);
    	if (value != null)
    		value.setValue(o, getModule().getField(index));
    }

    /**
     * Sets all values to empty
     */
    public void clearValues() {
        Set keys = values.keySet();
        for (Iterator iter = keys.iterator(); iter.hasNext(); ) {
            Integer key = (Integer) iter.next();
            DcValue value = (DcValue) values.get(key);
            value.setValue(null, getModule().getField(key.intValue()));;
        }
    } 

    /**
     * Retrieves the maximum length for the given field 
     * @param index field index
     */
    public int getMaxFieldLength(int index) {
        return getField(index).getMaximumLength();
    }
    
    /**
     * Returns the value of a field
     *
     * @param iField    field indicator
     * @return (String) field value
     */
    public Object getValue(int index) {
        return getValueDef(index).getValue();
    } 

    public String getDisplayString(int index) {
    	String s = "";
        if (getFieldType(index) != ComponentFactory._PICTUREFIELD) 
        	s = getValueDef(index).getDisplayString(getFieldType(index));
        
        return s;
    }
    
    public void setRow(int row) {
        uiRowNumber = row;
    }
    
    public int getRow() {
    	return uiRowNumber;
    }
    
    /**
     * Returns the UI field type for the given field
     * @see net.sf.dc.console.ComponentFactory
     *
     * @return UI field type index
     */    
    public int getFieldType(int index) {
        return getField(index).getFieldType();
    }

    /**
     * Returns the database field name for the given field
     *
     * @param index field index
     */    
    public String getDatabaseFieldName(int index) {
        return getField(index).getDatabaseFieldName();
    }

    /**
     * Inserts a new item. Requests to the UI can be passed.
     */
    public void saveNew() throws ValidationException {
    	markAsChanged();
        checkIntegrity(false);
        WorkFlow.insertValues(this);
    }
    
    /**
     * Updates an item. Requests to the UI can be passed.
     */
    public void saveUpdate() throws ValidationException {
        this.checkIntegrity(true);
        WorkFlow.updateValues(this);
    }
    
    public void delete() {
        WorkFlow.deleteValues(this);
    }

    /**
     * Loads items. Requests to the UI can be passed.
     * @param options query options, can be null
     * @param requestors  collection of requests (can be null)
     */
    public void load(QueryOptions options) {
        WorkFlow.searchValues(this, options);
    }

    /**
     * Disables the output of this query to the log and log screen.
     * Also disables showing messageboxes when a query was success-
     * fully executed
     *
     * @param silence show output y/n
     */
    public void setSilence(boolean silence) {
        this.bSilence = silence;
    }

    /**
     * Indicates if queries should display output
     */
    public boolean getSilence() {
        return this.bSilence;
    }

    /**
     * Specifies if values should be validated
     * @param validate validate values y/n
     */
    public void setValidation(boolean validate) {
        bValidation = validate;
    }

    /**
     * Returns the field definition object for the given field index
     * @param index field index
     */
    public DcField getField(int index) {
        return getModule().getField(index);
    }
    
    /**
     * Returns the labels of the fields that have been marked as changed
     */
    public String[] getChangedFields()  {
        ArrayList changedFields = new ArrayList();
        int counter = 0;
        for (Iterator iter = values.values().iterator(); iter.hasNext(); counter++) {
            DcValue value = (DcValue) iter.next();
            if (value.isChanged()) {
                changedFields.add(getField(counter).getDatabaseFieldName());
            }
        }
        String[] fieldNames = new String[changedFields.size()];
        return (String[]) changedFields.toArray(fieldNames);
    }
    
    /**
     * Checks if the values are unique and if all the required fields
     * have been filled
     * @param update indicates if the values are new or existing
     */
    public void checkIntegrity(boolean update) throws ValidationException {
        if (DcSettings.getValueAsBoolean(Repository.settings.stCheckRequiredFields)) {
            validateRequiredFields();
        }
        if (DcSettings.getValueAsBoolean(Repository.settings.stCheckUniqueness)) {
            isUnique(this, update);
        }
    }

    /**
     * Checks to see whether the item already exists or not
     *
     * @param o object to be checked
     * @param bUpdateQuery indicates whether the object is to 
     *        be updated or if it is a new object
     */
    public void isUnique(DcObject o, boolean bUpdateQuery) throws ValidationException {
        boolean bUnique = WorkFlow.checkUniqueness(o, bUpdateQuery);
        if (!bUnique && bValidation) {
            String message;
            if (getRow() != -1) {
                // shown by the table panels (ui row ID is set)
                message = DcResources.getText(Repository.messages.msgItemNotUniqueRow,
                          new String[] {toString(), String.valueOf(uiRowNumber + 1)});
            } else {
                // shown by item forms and the likes
                message = DcResources.getText(Repository.messages.msgItemNotUnique);
            }            
            
        	throw new ValidationException(message);
        }
    }

    /**
     * Checks to see whether al required fields have been filled
     */
    private String getUnfilledRequiredFields() {
        String required = "";
        Collection fields = getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.isRequired()) {
                Object value = getValue(field.getIndex());
                
                if (value == null || value.toString().equals("")) {
                    if (required.length() > 0) 
                        required += ",";
                    
                    required += getLabel(field.getIndex());
                }
            }
        }
        return required;
    }

    /**
     * Returns all field indices
     */
    public int[] getFieldIndices() {
        return getModule().getFieldIndices();
    }
    
    /**
     * Sets the module index / sets the ownership of this
     * data crow object  
     */
    protected abstract void setModuleIndex();
    
    /**
     * Initialize all user settings on this data crow object 
     */
    public void applySettings(DcFieldDefinitions definitions) {
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            DcFieldDefinition definition = (DcFieldDefinition) iter.next();
            DcField field = getField(definition.getIndex());
            
            String label = definition.getLabel();
            if (label != null && label.trim().length() > 0)
                field.setLabel(label);
            else 
            	field.setLabel(field.getSystemName());
            
            field.setEnabled(definition.isEnabled());
            field.setRequired(definition.isRequired());
            field.setHidden(!definition.isVisible());
        }
    }
    
    /**
     * Apply the IDs on this object and its children
     */
    public void setIDs() {
        String sID = Utilities.getUniqueID().substring(0, 15);
        setValue(DcObject._ID, sID);
        for (Iterator iter = alChildren.iterator(); iter.hasNext(); ) {
            DcObject child = (DcObject) iter.next();
            child.setIDs();
            child.setValue(child.getParentReferenceFieldIndex(), sID);
        }
    }    
    
    /**
     * Retrieves the value for the given field index
     * @param index field index
     */
    protected DcValue getValueDef(int index) {
        return (DcValue) values.get(new Integer(index));    
    }
    
    private void initializeSystemValues() {
        values.put(new Integer(DcObject._SYS_MODULE), new DcValue());
        setValue(DcObject._SYS_MODULE, getModule());
    }
    
    /**
     * Creates the value objects (empty value objects)
     */
    private void initializeValues() {
        int[] fields = getModule().getFieldIndices();
        for (int i = 0; i < fields.length; i++) {
            values.put(new Integer(fields[i]), new DcValue());
        }
    }    
    
    /**
     * Check if all required fields have been filled
     */
    private void validateRequiredFields() throws ValidationException {
        String s = this.getUnfilledRequiredFields();

        if (s.trim().length() > 1) {
            String message;
            if (getRow() != -1) {
            	//shown by the table panels (ui row ID is set)
                message = DcResources.getText(Repository.messages.msgIncompleteRecRow,
                		  new String[] {s, toString(), String.valueOf(getRow() + 1)});
            } else {
                //shown by item forms and the likes
                message = DcResources.getText(Repository.messages.msgIncompleteRec, s);
            }
            throw new ValidationException(message);
        }
    }
    
    public void finalize() throws Throwable {
        try {
            values = null;
            requests = null; 
            alChildren = null;
        } finally {
            super.finalize();
        }
    }
}
