/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.db;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMediaObject;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.wf.requests.RequestCollection;

/**
 * Creates a query String by using the QueryOptions, the QueryWhat, the 
 * QueryConditions and the query type
 *
 * @author Robert Jan van der Waals
 * @since 1.3
 * @version 0.5
 */
public class Query {

    public static final int _UNDEFINED = 0;
    public static final int _UPDATE = 1;
    public static final int _CREATE = 2;
    public static final int _INSERT = 3;
    public static final int _SELECT = 4;
    public static final int _DELETE = 5;

    private static final int _SELECTPRECISE = 6;
    private static final int _SELECTOR = 7;
    private static final int _SELECTPRECISEOR = 8;

    private RequestCollection requestors;
    private int queryType;

    private boolean bPreciseSelect = false;
    private boolean bComplyToAllConditions = true;

    //private DcObject dco;
    private String[] queries;
    private String objectID;
    
    private String orderBy;
    private boolean bSilence = false;
    private boolean isBatch = false; 
    private boolean endOfBatch = false;
    
    /**
     * Creates a new Query object. In this object the condition values and
     * their fields are stored, the columns of the object, the options, the
     * query type and the requestors of the query (which might want the query
     * to result in an action)
     * The public query types are not all there is to it, depending on the
     * query options, a subtype is set (only in case of select queries)
     * From these values a query string is created
     */
    public Query(   int queryType,   
                    DcObject dco,
                    QueryOptions options,
                    RequestCollection requestors) {
        
        if (options != null) {
            bPreciseSelect = options.getPreciseSelect();
            bComplyToAllConditions = options.getComplyAllConditions();
            orderBy = options.getOrderByField();
        }

        this.requestors = requestors;
        this.queryType = queryType;
        this.objectID = dco.getID();

        if (this.queryType == _SELECT) {
            if (bPreciseSelect && !bComplyToAllConditions) {
                this.queryType = _SELECTPRECISEOR;
            } else if (bPreciseSelect) {
                this.queryType = _SELECTPRECISE;
            } else if (!bComplyToAllConditions) {
                this.queryType = _SELECTOR;
            }
        }
        
        queries = getQueries(dco);
    }
    
    /**
     * Sets this query as being part of a batch
     * @param endOfBatch last of the batch y/n
     */
    public void setBatch(boolean endOfBatch) { 
        this.isBatch = true; 
        this.endOfBatch = endOfBatch;
    }
    
    /**
     * Indicates if this query is part of a batch
     */
    public boolean isBatch() {
        return isBatch;
    }

    /**
     * Indicates if this query is the end of a batch
     */
    public boolean isEndOfBatch() {
        return isBatch() && endOfBatch;
    }

    public String getObjectID() {
        return objectID;
    }
    
    /**
     * Returns the query type
     */
    public int getQueryType() {
        return queryType;
    }

    /**
     * Specifies whether the query should report events to the log
     * and the UI
     */
    public boolean getSilence() {
        return bSilence;
    }

    /**
     * Sets the query to silent. Silent means that no events are logged
     * @param silence  specifies if the query should run silently
     */
    public void setSilence(boolean silence) {
        this.bSilence = silence;
    }

    /**
     * Returns the requestors
     */
    public RequestCollection getRequestors() {
        return requestors;
    }
    
    public String[] getQueries() {
        return queries;
    }

    /**
     * Returns the query string. The query string is created if it doesn't 
     * exist as yet
     */
    private String[] getQueries(DcObject dco) {
        String[] queries = new String[1];
        switch (queryType) {
            case _UPDATE :
                queries[0] = getUpdateQuery(dco);
                break;
            case _CREATE :
                queries[0] = getCreateTableQuery(dco);
                break;
            case _INSERT :
                queries[0] = getInsertQuery(dco);
                break;
            case _DELETE :
                queries[0] = getDeleteQuery(dco);
                break;
            case _UNDEFINED :
            case _SELECT :
            case _SELECTPRECISE :
            case _SELECTOR :
            case _SELECTPRECISEOR :
                queries = getSelectQueries(dco);
                break;
        }
        
        return queries;
    }

    private boolean isEmpty(String s) {
        String emptyValue = DcResources.getText(Repository.labels.lblEmptySearchValue);
        return s.equals("''") || s.equals("'" + emptyValue + "'") || 
               s.equals(emptyValue) || s.equals("0");
    }
    
    private String getEmptyCondition(String column, String s) {
        String emptyCondition;
        if (s.equals("0")) {
            emptyCondition = "(" + column + " = 0 OR " + column + " IS NULL) ";
        } else {
            emptyCondition = "(" + column + " IS NULL OR " + column + " = '') ";
        }
        return emptyCondition;
    }
    
    private String getQueryValue(DcObject object, int index) {
        Object value = object.getValue(index);      
        if (value instanceof DcObject) 
            value = ((DcObject) value).getID();

        return value != null ? value.toString() : null;
    }
    
    private String getInsertQuery(DcObject dco) {
        StringBuffer values = new StringBuffer();
        StringBuffer columns = new StringBuffer();
        
        Collection fields = dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();

            if (dco.isChanged(field.getIndex()) && !field.isUiOnly()) {
                if (values.length() > 0)
                    values.append(", ");
                
                if (columns.length() > 0)
                    columns.append(", ");
                
                
                columns.append(field.getDatabaseFieldName());
                
                boolean integer = field.getValueType() == Repository.datatypes._INTEGER;
                if (!integer) values.append("'"); 
                values.append(getQueryValue(dco, field.getIndex()));
                if (!integer) values.append("'");
            }
        }
        
        return "INSERT INTO " + dco.getTableName() + " (" + columns + ") \r\n" +
               "VALUES (" + values + ");";
    }
    
    private String getCreateTableQuery(DcObject dco) {
        String columns = "";
        Collection fields = dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            
            if (!field.isUiOnly()) {
                if (columns.length() > 0) 
                    columns += ", ";
                
                columns += field.getDatabaseFieldName() + " " + field.getDataBaseFieldType();
                if (field.getIndex() == DcObject._ID) {
                    columns += " PRIMARY KEY";
                }
            }
        }
        
        return "CREATE CACHED TABLE " + dco.getTableName() + "\r\n" + 
               "(" + columns + ");";
    }
    
    private String getUpdateQuery(DcObject dco) {
        StringBuffer values = new StringBuffer();
        Collection fields = dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();

            if (dco.isChanged(field.getIndex()) && !field.isUiOnly()) {
                if (values.length() > 0)
                    values.append(", ");
                
                values.append("\r\n");
                values.append(field.getDatabaseFieldName());
                values.append(" = ");
                
                boolean integer = field.getValueType() == Repository.datatypes._INTEGER;
                
                if (!integer) values.append("'"); 
                values.append(getQueryValue(dco, field.getIndex()));
                if (!integer) values.append("'");
            }
        }
        
        return "UPDATE " + dco.getTableName() + " SET " + values + "\r\n" +
               "WHERE ID = '" + dco.getID() + "';";
    }
    
    private String getDeleteQuery(DcObject dco) {
        String query = "DELETE FROM " + dco.getTableName() + " WHERE ID = '" + dco.getID() + "';";
        return query;
    }
    
    private String[] getSelectQueries(DcObject dco) {
        String[] queries;
        int tableCount = 0;
        Collection tables = new ArrayList();
        if (dco.getModule().getIndex() == DCModules._GENERIC) {
            for (Iterator iter = DCModules.getModules(); iter.hasNext(); ) {
                DCModule module = (DCModule) iter.next();
                if (module.getIndex() != DCModules._BOOKCAT && module.getIndex() != DCModules._SOFTWARECAT && module.getIndex() != DCModules._GENERIC) {
                    tables.add(module.getTableName());
                    tableCount++;
                }
            }
        } else {
            tables.add(dco.getTableJoin());
            tableCount++;
        }

        queries = new String[tableCount];
        StringBuffer columns = new StringBuffer();
        
        int counter = 0;

        Collection objects = new ArrayList();
        objects.add(dco);
        objects.addAll(dco.getChildren());
        
        int counterTables = 0;
        for (Iterator iterTables = tables.iterator(); iterTables.hasNext(); counterTables++) {        
            String query = "";
            
            String table = (String) iterTables.next();
            StringBuffer conditions = new StringBuffer();
            
            for (Iterator iter = objects.iterator(); iter.hasNext(); counter++) {
                DcObject object = (DcObject) iter.next();
                Collection fields = object.getFields();
                
                for (Iterator iterFields = fields.iterator(); iterFields.hasNext(); ) {
                    DcField field = (DcField) iterFields.next();
                    
                    if (!field.isUiOnly()) {
                        boolean integer = field.getValueType() == Repository.datatypes._INTEGER;
                        
                        String column = field.getDatabaseFieldName();
                        if (objects.size() > 1) 
                            column = object.getTableShortName() + "." + column;
                        
                        if (counter == 0) {
                            if (columns.length() > 0)
                                columns.append(", ");
                            
                            // only select fields for the current module, child fields are added to the condition only
                            columns.append(column);
                        }
                        
                        if (object.isChanged(field.getIndex())) {
                            String value = getQueryValue(object, field.getIndex());
                            String condition = "";
                            if (isEmpty(value)) { 
                                condition = getEmptyCondition(column, value);
                            } else { 
                                String appender = queryType == _SELECTOR || queryType == _SELECTPRECISEOR ? " OR " : " AND ";
                                
                                if (conditions.length() > 0)
                                    conditions.append(appender);
                                
                                if (queryType == _SELECT || queryType == _SELECTOR) {
                                    condition = "UPPER (" + column + ") LIKE '%" + value.toUpperCase() + "%'";
                                } else {
                                    condition = "UPPER (" + column + ")= "; 
                                    if (!integer) condition += "'";
                                    condition += value.toUpperCase();
                                    if (!integer) condition += "'";
                                }
                            }
                            conditions.append(condition);
                        }
                    }
                }
            }
            
            query+= "SELECT DISTINCT " + columns + " FROM " + table + "\r\n";
            if (conditions.length() > 0) 
                query +=  "WHERE " + conditions;
        
            if (orderBy != null && orderBy.length() > 0) {
                if (objects.size() > 1) 
                    orderBy = dco.getTableShortName() + "." + orderBy;
    
                query += " ORDER BY " + orderBy;
            } else {
                if (objects.size() > 1) 
                    orderBy = dco.getTableShortName() + "." + orderBy;
                
                if (dco instanceof DcMediaObject)
                    orderBy += dco.getField(DcMediaObject._A_TITLE).getDatabaseFieldName();    
                else
                    orderBy += dco.getField(DcObject._ID).getDatabaseFieldName();
                
                query += " ORDER BY " + orderBy; 
            }
            
            query += " LIMIT " + DcSettings.getValueAsInt(Repository.settings.stMaximumSearchRes) + ";";
        
            queries[counterTables] = query;
        }
        
        return queries;
    }
    
    public void finalize() throws Throwable {
        try {
            queries = null;
            objectID = null;
            requestors = null;
            orderBy = null;
        } finally {
            super.finalize();
        }
    }    
}
