/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.http;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;

import net.sf.dc.core.Repository;
import net.sf.dc.settings.DcSettings;
import net.sf.ext.util.Base64;

/**
 * HttpURLConnection implementation with proxy implementation
 * 
 * @author Robert Jan van der Waals
 * @since 1.4
 * @version 1.9
 */
public class HttpConnection {
    private final HttpURLConnection uc;
    
    public HttpConnection(URL url) throws Exception {
        uc = getConnection(url); 
    }

    /**
     * Read the page from the given location to a String
     * @param url the target
     */
    public String getString() throws IOException {
        BufferedInputStream in = new BufferedInputStream(uc.getInputStream());       
        InputStreamReader reader = new InputStreamReader(in, "UTF-8");
        
        StringBuffer sb = new StringBuffer();
        int c;
        while ((c = reader.read()) != -1) {
            sb.append((char) c);
        }    
        reader.close();
        in.close();
        uc.disconnect();
        return sb.toString();
    }
    
    public int getContentLength() {
    	return uc.getContentLength();
    }
    
    public void setContentLength(int length) {
        uc.setRequestProperty("Content-Length", String.valueOf(length));
    }
    
    public void setToSoapRequest(String soapAction) throws Exception {
        uc.setRequestProperty("Content-Type", "text/xml; charset=utf-8");
        uc.setRequestProperty("SOAPAction", soapAction);
        uc.setRequestMethod("POST");   
        uc.setDoOutput(true);
        uc.setDoInput(true);        
    }
    
    public OutputStream getOutputStream() throws IOException {
        uc.setDoOutput(true);
        return uc.getOutputStream();        
    }
    
    /**
     * Read the bytes from the given location to a String.
     * Used for images.
     * @param url the target
     */    
    public final byte[] getBytes() throws IOException {
        int contentLength = uc.getContentLength();
        
        InputStream raw = uc.getInputStream();
        BufferedInputStream in  = new BufferedInputStream(raw);
        byte[] data = new byte[contentLength];
        int bytesRead = 0;
        int offset = 0;
        while (offset < contentLength) {
            bytesRead = in.read(data, offset, data.length-offset);
            if (bytesRead == -1) {
                break;
            }
            offset += bytesRead;
        }
        in.close();
        
        if (offset != contentLength) {
            throw new IOException("Only read " + offset + " bytes; Expected " + 
                                  contentLength + " bytes");
        }     
        uc.disconnect();
        return data;
    }
    
    /**
     * Retrieves a picture from the net and returns this as a BASE64 string
     * @param url target url
     */
    public final String getImageAsBase64() throws IOException {
        byte[] bytes = getBytes();
        String image = String.valueOf(Base64.encode(bytes));
        return image.length() < 1200 ? "" : image;   
    }
    
    public void close() {
    	uc.disconnect();
    }
    
    /**
     * Get a Http connection with the correct proxy settings
     * @param url the target url
     */
    private final HttpURLConnection getConnection(URL url) throws IOException {
        String proxy = DcSettings.getValueAsString(Repository.settings.stProxyServerName);
        
        String username = DcSettings.getValueAsString(Repository.settings.stProxyUserName);
        String password = DcSettings.getValueAsString(Repository.settings.stProxyPassword);        
        int proxyPort = DcSettings.getValueAsInt(Repository.settings.stProxyServerPort);
        
        boolean useProxy = proxy.length() > 0 && proxyPort > 0;
        if (useProxy) {
        	System.setProperty("proxySet", "true");
        	System.setProperty("http.proxyHost", proxy);
        	System.setProperty("http.proxyPort", "" + proxyPort);
        } else {
            System.setProperty("proxySet", "false");
            System.setProperty("http.proxyHost", "");
            System.setProperty("http.proxyPort", "" + -1);
        }
        
        HttpURLConnection uc = (HttpURLConnection) url.openConnection();
        if (useProxy && username.trim().length() > 0) {
    		String proxyUPB64 = Base64.encode(username + ":" + password);
    		uc.setRequestProperty("Proxy-Authorization", "Basic " + proxyUPB64);
        }
        
        return uc;
    }    
}
