/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.processes.http;

import java.net.URL;
import java.util.Vector;

import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.console.forms.InternetSearchForm;
import net.sf.dc.core.Repository;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.movie.Movie;
import net.sf.dc.http.HttpConnection;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.ext.util.StringUtils;

/**
 * Access www.imdb.com to gather movie specific information
 *
 * @author Robert Jan van der Waals
 * @since 1.4
 * @version 1.6
 */
public class IMDBMovieUpdate extends InternetUpdateProcess {

    private final InternetSearchForm ui;
    private String searchString;
    private String server;
    
    private final static int _DIRECTEDBY = 0;
    private final static int _RUNTIME = 1;
    private final static int _USERRATING = 2;
    private final static int _COUNTRY = 3;
    private final static int _CERTIFICATION = 4;
    private final static int _GENRE = 5;
    
    private String[][] headers = {{"Directed by", "Dirigida por", "Ralis par", "Regie", "Regia di"},
                                  {"Runtime", "Duraci&oacute;n:", "Dure:", "L&auml;nge:", "Durata:"},
                                  {"User Rating:", "Calificaci&oacute;n de los usuarios:", "Note des utilisateurs:", "Ihre Bewertung:", "Voti degli utenti:"},
                                  {"Country:", "Pa&iacute;s:", "Pays:", "Produktionsland:", "Nazionalit:"},
                                  {"Certification", "Clasificaci&oacute;n:", "Classification:", "Altersfreigabe:", "Divieti:"},
                                  {"Genre", "Genere:"}};

    /**
     * initializes the IMDB query
     * @param searchString the string on which will be searched
     */
    public IMDBMovieUpdate(InternetSearchForm ui, String searchString) {
        this.ui = ui;
        this.searchString = searchString.replace(' ', '+');
    } 
    
    public void run() {
        server = DcSettings.getValueAsString(Repository.settings.stImdbServer);
    	String[] IDs = getIDs();
        ui.initProgressBar(IDs.length);

        if (IDs.length == 0) {
            new MessageBox(DcResources.getText(Repository.messages.msgNoResultsForKeywords,
                           searchString), MessageBox._WARNING);
            return;
        }
        
        ui.addMessage(DcResources.getText(Repository.messages.msgFoundXResults, 
                      String.valueOf(IDs.length)));      
        ui.addMessage(DcResources.getText(Repository.messages.msgStartParsingXResults, 
                      String.valueOf(IDs.length)));          
        
        for (int i = 0; i < IDs.length && !cancelled; i++) {
            try {
            	DcObject dco = parseInformation(IDs[i]);
                ui.addMessage(DcResources.getText(Repository.messages.msgParsingSuccessfull, 
                              dco.getValue(Movie._A_TITLE).toString()));          
                ui.updateProgressBar(i + 1);
            	ui.addObject(dco);
            } catch (Exception exp) {
            	ui.addMessage(DcResources.getText(Repository.messages.msgParsingError, exp.getMessage()));
                Messages.add(exp);
                ui.updateProgressBar(i + 1);
            }
        }
        ui.setSearchMode();
    }

    /**
     * Searches for the maching ID's from the search result(s)
     */
    public String[] getIDs() {
        Vector vKeys = new Vector(0);
        
        try {
            searchString = searchString.replaceAll(" ", "%20");
            ui.addMessage(DcResources.getText(Repository.messages.msgConnectingToServer, server));
            
            URL url = new URL(server + "/Find?" + searchString);
    
            HttpConnection connection = new HttpConnection(url);
            String sWebPage = connection.getString();
            connection.close();
    
            String sTemp = sWebPage;
            if (sTemp.indexOf("/title/tt") > -1) {
                int temp = 0;
                while (temp != -1) {
                    temp = sTemp.indexOf("/title/tt");
                    if (temp != -1) {
                        sTemp = sTemp.substring(temp + 9, sTemp.length());
                        String key  = sTemp.substring(0, 7);
                        if (!vKeys.contains(key)) {
                        	vKeys.add(key);
                        }
                    }
                }
            }
        } catch (Exception e) {
            new MessageBox(DcResources.getText(Repository.messages.msgCouldNotConnectToImdb), MessageBox._ERROR);
            Messages.add(e);
            ui.addMessage(DcResources.getText(Repository.messages.msgCouldNotConnectToImdb));
        }

        String[] keys = new String[vKeys.size()];
        return (String[]) vKeys.toArray(keys);
    }
    
    private int indexOf(int headerindex, String source) {
        int index = -1;
        for (int i = 0; i < headers[headerindex].length; i++) {
            index = source.indexOf(headers[headerindex][i]);
            if (index > -1) break;
        }
        return index;
    }
    
    private String getPart(int headerindex, String end, String source) {
        String s = "";
        for (int i = 0; i < headers[headerindex].length; i++) {
            s = StringUtils.getValueBetweenStrings(headers[headerindex][i], end, source);
            if (s != null && s.length() > 1) break;
        }
        return s;
    }
    
    /**
     * Reads the information from imdb.com
     * @param sID the movie ID from imdb.com
     */
    public Movie parseInformation(String sID) throws Exception {
        Movie movie = new Movie();
        
        movie.setValue(Movie._G_WEBPAGE, server + "/title/tt" + sID);
        movie.setValue(Movie._B_DESCRIPTION, getDescription(sID));

        URL url = new URL("http://www.imdb.com/title/tt" + sID);
        HttpConnection connection = new HttpConnection(url);
        String sWebPage = connection.getString();
        connection.close();
        
        String sTitle = StringUtils.getValueBetweenStrings("<title>", "</title>", sWebPage);
        movie.setValue(Movie._A_TITLE, StringUtils.convertHttpString(sTitle));
        movie.setValue(Movie._C_YEAR,  StringUtils.getValueBetweenStrings("(", ")", sTitle));
 
        
        String sDirected = getPart(_DIRECTEDBY, "</table>", sWebPage);
        sDirected = StringUtils.getValueBetweenStrings("/" + '"' + ">", "</a><br>", sDirected);
        movie.setValue(Movie._J_DIRECTOR, StringUtils.convertHttpString(sDirected));

        String sGenre = getPart(_GENRE, "<br>", sWebPage);
        int iGenre = sGenre.indexOf("/\">");
        String s = "";
        while (iGenre != -1) {
            sGenre = sGenre.substring(iGenre + 3);
            s += sGenre.substring(0, sGenre.indexOf("<")) + (",");
            iGenre = sGenre.indexOf("/\">");
        }
        sGenre = s.length() > 1 ? s.substring(0, s.length() -1) : s;
        movie.setValue(Movie._H_GENRE, StringUtils.convertHttpString(sGenre));
            
            
        int iRuntime = indexOf(_RUNTIME, sWebPage);
        String sRuntime = "";
        if (iRuntime != -1) {
            String sTemp = sWebPage.substring(iRuntime + 13);
            if (sTemp.indexOf("<") < sTemp.indexOf(" ")) {
                sRuntime = sTemp.substring(0, sTemp.indexOf("<"));
            } else {
                iRuntime = sTemp.indexOf("\t");
                if (iRuntime != -1) {
                    if (sTemp.indexOf(" ") < iRuntime) {
                        sRuntime = sTemp.substring(0, sTemp.indexOf(" "));
                    } else {
                        sRuntime = sTemp.substring(0, sTemp.indexOf("\t"));
                    }
                } else {
                     sRuntime = sTemp.substring(0, sTemp.indexOf(" "));
                }
            }
            try {
                iRuntime = new Integer(sRuntime).intValue() * 60;
                sRuntime = "" + iRuntime;
            } catch (NumberFormatException e) {
                sRuntime = "";
            }
        }
        
        movie.setValue(Movie._L_PLAYLENGTH, sRuntime);
        
        int iUserRating = indexOf(_USERRATING, sWebPage);
        if (iUserRating != -1) {
            String sTemp = sWebPage.substring(iUserRating + 20);
            if (sTemp.indexOf("<b>") > -1 && sTemp.indexOf("</b>") > -1) {
                try {
                    String vote = sTemp.substring(sTemp.indexOf("<b>") + 3, sTemp.indexOf("</b>"));
                	vote = vote.substring(0, vote.indexOf("/"));
                    int value = Math.round(Float.valueOf(vote).floatValue() / 2);
                    movie.setValue(Movie._E_RATING, new Integer(value));
                } catch (Exception exp) {}
            } 
        }        
        
        String sCountry = getPart(_COUNTRY, "/a>", sWebPage);
        sCountry = StringUtils.getValueBetweenStrings("/" + '"' + '>', "<", sCountry);
        movie.setValue(Movie._K_COUNTRY, StringUtils.convertHttpString(sCountry)); 

        String sCertification = getPart(_CERTIFICATION, "<br>", sWebPage);
        int iCertification = sCertification.indexOf("\">");
        s = "";
        while (iCertification != -1) {
            sCertification = sCertification.substring(iCertification + 2);
            s += sCertification.substring(0, sCertification.indexOf("<")) + (",");
            iCertification = sCertification.indexOf("\">");
        }
        sCertification = s.length() > 1 ? s.substring(0, s.length() -1) : s;
        movie.setValue(Movie._3_CERTIFICATION, StringUtils.convertHttpString(sCertification));        
        
        int iActors = sWebPage.indexOf("ast overview");
        if (iActors == -1)
            iActors = sWebPage.indexOf("cast:");
        
        if (iActors != -1) {
            String sTemp = sWebPage.substring(iActors);
            iActors = sTemp.indexOf("href=\"");
            sTemp = sTemp.substring(iActors, sTemp.indexOf("</table>"));
            iActors = sTemp.indexOf("(more)");
            if (iActors != -1) {
                sTemp = sTemp.substring(0, iActors - 50);
            }

            iActors = sTemp.indexOf("\">");
            String actors = "";
            while (iActors != -1) {
                try {
                    sTemp = sTemp.substring(iActors + 2);
                    actors += sTemp.substring(0, sTemp.indexOf("<")) + (", ");
                    
                    iActors = sTemp.indexOf("\"top\">");
                    if (iActors > -1) {
                        sTemp = sTemp.substring(iActors + 6);
                    }
                    
                    iActors = sTemp.indexOf("<");
                    if (iActors > 2) {
                        actors += sTemp.substring(0, sTemp.indexOf("<") - 1) + (", ");
                    }
                    
                    iActors = sTemp.indexOf("href");
                    if (iActors > -1) {
                        sTemp = sTemp.substring(iActors + 4);
                    }
                } catch (Exception exp) {
                    Messages.add(exp);
                }
                
                iActors = sTemp.indexOf("\">");
            }
            
            movie.setValue(Movie._I_ACTORS, 
                           StringUtils.convertHttpString((actors.substring(0, actors.length() - 1))));
        }
        
        String poster = getPoster(sID);
        if (poster == null) {
            poster =  getThumbnailImage(sWebPage);  
        }
        
        movie.setValue(Movie._X_PICTUREFRONT, poster);
        
        return movie;
    }
    
    private String getThumbnailImage(String webpage) {
        String poster = "";

        int iThumb = webpage.indexOf("<a name=\"poster\"");
        if (iThumb !=-1){
            String sTemp = webpage.substring(iThumb);
    
            String picture =
            StringUtils.getValueBetweenStrings("src=\"" , "\"", sTemp);
    
            try {
                URL url = new URL(picture);
        
                HttpConnection connection = new HttpConnection(url);
                poster = connection.getImageAsBase64();
            } catch (Exception exp) {
                Messages.add(exp);
            }
        }
        return poster;
    }

    /**
     * Gets the description for this movie
     * @param sID imdb movie ID
     */
    public String getDescription(String sID) {
        try {
            URL url = new URL(server + "/Plot?" + sID);
            HttpConnection connection = new HttpConnection(url);
            String sPlot = connection.getString();
            connection.close();
            
            String s = "<p class=" + '"' + "plotpar" + '"' + ">";
            int index = sPlot.indexOf(s);
            if (index > -1) {
                sPlot = StringUtils.getValueBetweenStrings(s, "</p>", sPlot);
                return StringUtils.convertHttpString(sPlot);
            } 
        } catch (Exception ignore) {}
        return "";
    }    
    
    /**
     * Retrieves the movie poster from imdb.com
     * @param sID the imdb movie ID
     */
    private String getPoster(String sID) {
        try {
            URL url = new URL(server + "/title/tt" + sID + "/posters");
            HttpConnection connection = new HttpConnection(url);
            String picture = connection.getString();
            connection.close();
            
            int iPicture = picture.indexOf("Poster images");
            if (iPicture > -1) {
                picture = picture.substring(iPicture);
                picture = StringUtils.getValueBetweenStrings("background=" + '"', '"' +">", picture);
                
                String picUrl = "posters.imdb.com";
                int iPicture2 = picture.indexOf(picUrl);
                if (iPicture2 != -1) {
                    String picUrl2 = picture.substring(iPicture2 + picUrl.length());
                    url = new URL("http://posters.imdb.com" + picUrl2);
                    connection = new HttpConnection(url);
                    String image = connection.getImageAsBase64();
                    
        	    	return image.length() < 1200 ? null : image;
                } 
            }
        } catch (Exception exp) {
            ui.addMessage(DcResources.getText(Repository.messages.msgPictureCouldNotBeRetrieved));
            Messages.add(exp);
        }
        
        return null;
    }
}
