/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.wf;

import java.sql.ResultSet;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Collection;

import javax.swing.SwingUtilities;

import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.generic.GenericObject;
import net.sf.dc.db.Database;
import net.sf.dc.db.Query;
import net.sf.dc.db.QueryOptions;
import net.sf.dc.messages.Messages;
import net.sf.dc.wf.requests.RequestCollection;
import net.sf.ext.util.Base64;

/**
 * This is the layer between the UI and the database. Checks for data
 * integrity and checks on bussiness rules are done here
 *
 * @author Robert Jan van der Waals
 * @since 1.3
 * @version 2.4
 */
public abstract class WorkFlow {

    private static Database db;

    /**
     * starts the database if not running
     */
    public static void startDatabase() {
        if (db == null) {
            db = new Database();
        }
    }

    /**
     * Restarts the database. Old database will remain running to
     * complete previous tasks !
     */
    public static void restartDatabase() {
        db = null;
        startDatabase();
    }

    /**
     * Searches for values. The query is put in the query queue. The results are
     * past to the caller by using the RequestorCollection
     * @param o the object which is used as the condition
     * @param options query options
     */
    public static void searchValues(DcObject o, QueryOptions queryOptions) {
        startDatabase();
        db.searchValues(o, queryOptions);
    }

    /**
     * Inserts new values. The query is put in the query queue. Actions are
     * past to the caller by using the RequestorCollection
     * @param o the object to be inserted
     */
    public static void insertValues(DcObject o) {
        startDatabase();
        db.insertValues(o);
    }

    /**
     * Updates values. The query is put in the query queue. Actions are
     * past to the caller by using the RequestorCollection
     * @param o  the object to be updated
     */
    public static void updateValues(DcObject o) {
        startDatabase();
        db.updateValues(o);
    }

    /**
     * Deletes the object. The query is put in the query queue. Actions are
     * past to the caller by using the RequestorCollection
     * @param o the object to be updated
     */    
    public static void deleteValues(DcObject o) {
        startDatabase();
        db.deleteValues(o);
    }
    
    public static Collection runQueryDirect(String sQuery, boolean catchErrors) throws Exception{
        startDatabase();
        return db.runQueryDirect(sQuery, catchErrors);     
    }

    /**
     * Runs a query directly, bypassing the query queue
     * @param o Data Crow object
     */
    public static Collection runQueryDirect(DcObject o) {
        startDatabase();
        return db.runQueryDirect(o);
    }

    /**
     * Does not put the query in a queue but executes it immediately.
     * Errors are returned to the caller (not thrown)
     * @param Query Query object
     */
    public static Collection runQueryDirect(Query query) {
        startDatabase();
        return db.runQueryDirect(query);
    }

    /**
     * Does not put the query in a queue but executes it immediately.
     * Does not close the resultset!
     * @param sQuery query statement
     */
    public static ResultSet runQueryDirectUnclosed(String sQuery) throws Exception {
        startDatabase();
        return db.runQueryDirectUnclosed(sQuery, false);
    }

    /**
     * Does not put the query in a queue but executes it immediately.
     * @param sQuery query statement
     */
    public static Collection runQueryDirect(String sQuery) {
        startDatabase();
        return db.runQueryDirect(sQuery);
    }

    /**
     * Passes requests thru to the UI (<url>MainFrame</url>)
     * @param vTable data to be shown
     * @param requestorObjects collection of requests
     */
    public static void displayResults(Collection objects, RequestCollection requestorObjects, boolean qryWasSuccess) {
        Runnable runMe = new UIUpdater(objects, requestorObjects, qryWasSuccess);
        SwingUtilities.invokeLater(runMe);
        runMe = null;
    }
    
    

    /**
     * Checks if the object is unique
     * @param o object to be checked
     * @param bUpdate specifies if the object exists already in the
     *        database or not
     */
    public static boolean checkUniqueness(DcObject o, boolean bUpdate) {
        startDatabase();
        return db.uniqueValues(o, bUpdate);
    }

    /**
     * Converts a resultset to a DataCrowObject Vector
     * @param result result set from a query
     */
    public static Collection convertToDCObjects(ResultSet result) {
        ArrayList objects = new ArrayList();

        try {
        	result.isLast();
        } catch (Exception exp) {
            return objects;
        }
        
        try {
            int columnCount = result.getMetaData().getColumnCount();
            String[] columns = new String[columnCount];

            int k = 0;
            for (int i = 1; i < columnCount + 1; i++) {
                columns[k] = result.getMetaData().getColumnName(i);
                k++;
            }

            String sTableName = result.getMetaData().getTableName(1);
            
            while (result.next()) {
                DcObject dco = DCModules.getObjectForTable(sTableName);

                for (int i = 0; i < columns.length; i++) {
                    int type = result.getMetaData().getColumnType(i + 1);
                    String column = columns[i];
                    Object value = null;
                    
                    if (type == Types.BINARY) {
                        value = result.getBytes(column);
                        if (value != null) {
                            value = Base64.encode((byte[]) value);
                        } 
                    } else {
                        value = result.getString(column);    
                    }
                    dco.setValueForColumn(column, value);
                }
                
                dco.setValue(GenericObject._SYS_MODULE, dco.getModule());
                objects.add(dco);
            }
        } catch (Exception exp) {
            Messages.add(exp);
        } 
        return objects;
    }
}
