package net.sf.ext.mediamanager.ebm.matroska;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.ListIterator;

import net.sf.ext.mediamanager.ebml.BinaryElement;
import net.sf.ext.mediamanager.ebml.DataSource;
import net.sf.ext.mediamanager.ebml.EBMLReader;
import net.sf.ext.mediamanager.ebml.Element;
import net.sf.ext.mediamanager.ebml.FloatElement;
import net.sf.ext.mediamanager.ebml.InputStreamDataSource;
import net.sf.ext.mediamanager.ebml.MasterElement;
import net.sf.ext.mediamanager.ebml.SignedInteger;
import net.sf.ext.mediamanager.ebml.StringElement;
import net.sf.ext.mediamanager.ebml.UnsignedInteger;

/**
 * <p>Title: JEBML</p>
 * <p>Description: Java Classes to Read EBML Elements</p>
 * <p>Copyright: Copyright (c) 2004 Jory Stone <jcsston@toughguy.net></p>
 * @author jcsston
 * @version 1.0
 */

public class MatroskaFile {
  /**
   * Matroska Frame, holds a Matroska frame timecode, duration, and Data
   */
  public class MatroskaFrame {
    public int TrackNo;
    public long Timecode;
    public long Duration;
    public long Reference;
    public byte [] Data;

    /**
     * MatroskaFrame Default constructor
     */
    public MatroskaFrame() {
      //System.out.println("new " + this);
    }

    /**
     * MatroskaFrame Copy constructor
     * @param copy MatroskaFrame to copy
     */
    public MatroskaFrame(MatroskaFrame copy) {
      //System.out.println("MatroskaFrame copy " + this);
      this.TrackNo = copy.TrackNo;
      this.Timecode = copy.Timecode;
      this.Duration = copy.Duration;
      if (copy.Data != null) {
        this.Data = new byte[copy.Data.length];
        System.arraycopy(copy.Data, 0, this.Data, 0, copy.Data.length);
      }
    }
  }
  /**
   * Matroska Frame Puller interface
   */
  public interface MatroskaFramePuller {
    public void PushNewMatroskaFrame(MatroskaFrame frame);
  }

  /**
   * Matroska Track Class
   */
  public class MatroskaFileTrack {
    public short TrackNo;
    public long TrackUID;
    public byte TrackType;
    public long DefaultDuration;
    public String Name;
    public String Language;
    public String CodecID;
    public byte [] CodecPrivate;

    public short Video_PixelWidth;
    public short Video_PixelHeight;
    public short Video_DisplayWidth;
    public short Video_DisplayHeight;
    public float Audio_SamplingFrequency;
    public float Audio_OutputSamplingFrequency;
    public short Audio_Channels;
    public byte Audio_BitDepth;

    /**
     * Converts the Track to String form
     * @return String form of MatroskaFileTrack data
     */
    public String toString() {
      String s = new String();

      s += "\t\t" + "TrackNo: " + TrackNo + "\n";
      s += "\t\t" + "TrackUID: " + TrackUID + "\n";
      s += "\t\t" + "TrackType: " + MatroskaDocType.TrackTypeToString(TrackType) + "\n";
      s += "\t\t" + "DefaultDuration: " + DefaultDuration + "\n";
      s += "\t\t" + "Name: " + Name + "\n";
      s += "\t\t" + "Language: " + Language + "\n";
      s += "\t\t" + "CodecID: " + CodecID + "\n";
      if (CodecPrivate != null)
        s += "\t\t" + "CodecPrivate: " + CodecPrivate.length + " byte(s)" + "\n";

      if (TrackType == MatroskaDocType.track_video) {
        s += "\t\t" + "PixelWidth: " + Video_PixelWidth + "\n";
        s += "\t\t" + "PixelHeight: " + Video_PixelHeight + "\n";
        s += "\t\t" + "DisplayWidth: " + Video_DisplayWidth + "\n";
        s += "\t\t" + "DisplayHeight: " + Video_DisplayHeight + "\n";
      }

      if (TrackType == MatroskaDocType.track_audio) {
        s += "\t\t" + "SamplingFrequency: " + Audio_SamplingFrequency + "\n";
        if (Audio_OutputSamplingFrequency != 0)
          s += "\t\t" + "OutputSamplingFrequency: " + Audio_OutputSamplingFrequency + "\n";
        s += "\t\t" + "Channels: " + Audio_Channels + "\n";
        if (Audio_BitDepth != 0)
          s += "\t\t" + "BitDepth: " + Audio_BitDepth + "\n";
      }

      return s;
    }
  }
  public class MatroskaFileSimpleTag {
    public String Name;
    public String Value;
    public ArrayList Children = new ArrayList();

    public String toString(int depth) {
      String s = new String();
      String depthIndent = new String();
      for (int d = 0; d < depth; d++)
        depthIndent += "\t";

      s += depthIndent + "SimpleTag\n";
      s += depthIndent + "\tName: " + Name + "\n";
      s += depthIndent + "\tValue: " + Value + "\n";

      depth++;
      for (int t = 0; t < Children.size(); t++) {
        s += ((MatroskaFileSimpleTag)Children.get(t)).toString(depth);
      }

      return s;
    }
  }
  public class MatroskaFileTagEntry {
    public ArrayList TrackUID = new ArrayList();
    public ArrayList ChapterUID = new ArrayList();
    public ArrayList AttachmentUID = new ArrayList();
    public ArrayList SimpleTags = new ArrayList();

    public String toString() {
      String s = new String();

      if (TrackUID.size() > 0) {
        s += "\t\t" + "TrackUID: " + TrackUID.toArray().toString() + "\n";
      }
      if (ChapterUID.size() > 0) {
        s += "\t\t" + "ChapterUID: " + ChapterUID.toArray().toString() + "\n";
      }
      if (AttachmentUID.size() > 0) {
        s += "\t\t" + "AttachmentUID: " + AttachmentUID.toArray().toString() + "\n";
      }

      for (int t = 0; t < SimpleTags.size(); t++) {
        s += ((MatroskaFileSimpleTag)SimpleTags.get(t)).toString(2);
      }

      return s;
    }
  }

  /**
   * Number of Clusters to search before assuming that a track has ended
   */
  public static int CLUSTER_TRACK_SEARCH_COUNT = 4;

  protected DataSource ioDS;
  protected EBMLReader reader;
  protected Element level0 = null;
  protected String SegmentTitle;
  protected String MuxingApp;
  protected String WritingApp;
  protected long TimecodeScale = 1000000;
  protected double Duration;
  protected ArrayList TrackList = new ArrayList();
  protected ArrayList TagList = new ArrayList();
  protected LinkedList FrameQueue = new LinkedList();

  private String Language="";
  private String SubLang="";
  private String VidCodec="";
  private String AudCodec="";
  private String AudChannels="";
  private String AudRate="";
  private String Resolution="";
  
  /**
   * Primary Constructor for Matroska File class.
   *
   * @param inputStream InputStream to read the Matroska file from
   */
  public MatroskaFile(java.io.InputStream inputStream) {
    ioDS = new InputStreamDataSource(inputStream);
    reader = new EBMLReader(ioDS, MatroskaDocType.obj);
  }
  public MatroskaFile(DataSource inputDataSource) {
    ioDS = inputDataSource;
    reader = new EBMLReader(ioDS, MatroskaDocType.obj);
  }

  /**
   * Read / Parse the Matroska file.
   * @throws RuntimeException On various errors
   */
  public void readFile() {
    Element level1 = null;
    Element level2 = null;

    level0 = reader.readNextElement();
    if (level0 == null) {
      throw new java.lang.RuntimeException("Error: Unable to scan for EBML elements");
    }

    if (level0.equals(MatroskaDocType.EBMLHeader_Id)) {
      level1 = ((MasterElement)level0).readNextChild(reader);

      while (level1 != null) {
        level1.readData(ioDS);
        if (level1.equals(MatroskaDocType.DocType_Id)) {
          String DocType = ((StringElement)level1).getValue();
          if (DocType.compareTo("matroska") != 0) {
            throw new java.lang.RuntimeException("Error: DocType is not matroska, " + ((StringElement)level1).getValue());
          }
        }
        level1 = ((MasterElement)level0).readNextChild(reader);
      }
    } else {
      throw new java.lang.RuntimeException("Error: EBML Header not the first element in the file");
    }

    level0 = reader.readNextElement();
    if (level0.equals(MatroskaDocType.Segment_Id)) {
      level1 = ((MasterElement)level0).readNextChild(reader);

      while (level1 != null) {
        if (level1.equals(MatroskaDocType.SegmentInfo_Id)) {
          _parseSegmentInfo(level1, level2);

        } else if (level1.equals(MatroskaDocType.Tracks_Id)) {
          _parseTracks(level1, level2);

        } else if (level1.equals(MatroskaDocType.Cluster_Id)) {
          _parseNextCluster(level1);
          // Break out of this loop, we should only parse the first cluster
          break;

        } else if (level1.equals(MatroskaDocType.Tags_Id)) {
          _parseTags(level1, level2);

        }

        level1.skipData(ioDS);
        level1 = ((MasterElement)level0).readNextChild(reader);
      }
    } else {
      throw new java.lang.RuntimeException("Error: Segment not the second element in the file");
    }
  }
  /**
   * Get the Next MatroskaFrame
   *
   * @return The next MatroskaFrame in the queue, or null if the file has ended
   */
  public MatroskaFrame getNextFrame() {
    if (FrameQueue.isEmpty()) {
      _fillFrameQueue();
    }

    // If FrameQueue is still empty, must be the end of the file
    if (FrameQueue.isEmpty()) {
      return null;
    }
    return (MatroskaFrame)FrameQueue.removeFirst();
  }
  /**
   * Get the Next MatroskaFrame, limited by TrackNo
   *
   * @param TrackNo The track number to only get MatroskaFrames from
   * @return The next MatroskaFrame in the queue, or null if there are no more frames for the TrackNo track
   */
  public MatroskaFrame getNextFrame(int TrackNo) {
    if (FrameQueue.isEmpty()) {
      _fillFrameQueue();
    }

    // If FrameQueue is still empty, must be the end of the file
    if (FrameQueue.isEmpty()) {
      return null;
    }
    int tryCount = 0;
    MatroskaFrame frame = null;
    ListIterator iter = FrameQueue.listIterator();
    while (frame == null) {
      if (iter.hasNext()) {
        frame = (MatroskaFrame)iter.next();
        if (frame.TrackNo == TrackNo) {
          iter.remove();
          return frame;
        }
        frame = null;
      } else {
        _fillFrameQueue();
        // Update Iterator
        int index = iter.nextIndex();
        iter = FrameQueue.listIterator(index);
        if (++tryCount > CLUSTER_TRACK_SEARCH_COUNT) {
          // If we have not found any frames belonging to a track in 4 clusters
          // there is a good chance that the track is over
          return null;
        }
      }
    }
    return frame;
  }

  private void _fillFrameQueue() {
    if (level0 == null)
      throw new java.lang.IllegalStateException("Call readFile() before reading frames");

    Element level1 = ((MasterElement)level0).readNextChild(reader);
    if (level1 != null) {
      if (level1.equals(MatroskaDocType.Cluster_Id)) {
        _parseNextCluster(level1);
      }
      level1.skipData(ioDS);
    }
  }

  private void _parseNextCluster(Element level1) {
    Element level2 = null;
    Element level3 = null;
    long ClusterTimecode = 0;
    level2 = ((MasterElement)level1).readNextChild(reader);

    while (level2 != null) {
      if (level2.equals(MatroskaDocType.ClusterTimecode_Id)) {
        level2.readData(ioDS);
        ClusterTimecode = ((UnsignedInteger)level2).getValue();

      } else if (level2.equals(MatroskaDocType.ClusterBlockGroup_Id)) {
        MatroskaBlock block = null;
        long BlockDuration = 0;
        long BlockReference = 0;
        level3 = ((MasterElement)level2).readNextChild(reader);

        while (level3 != null) {
          if (level3.equals(MatroskaDocType.ClusterBlock_Id)) {
            block = (MatroskaBlock)level3;
            block.parseBlock(ioDS);

          } else if (level3.equals(MatroskaDocType.ClusterBlockDuration_Id)) {
            level3.readData(ioDS);
            BlockDuration = ((UnsignedInteger)level3).getValue();

          } else if (level3.equals(MatroskaDocType.ClusterReferenceBlock_Id)) {
            level3.readData(ioDS);
            BlockReference = ((SignedInteger)level3).getValue();
          }

          level3.skipData(ioDS);
          level3 = ((MasterElement)level2).readNextChild(reader);
        }

        if (block == null)
          throw new java.lang.NullPointerException("BlockGroup element with no child Block!");

        MatroskaFrame frame = new MatroskaFrame();
        frame.TrackNo = block.getTrackNo();
        frame.Timecode = block.getAdjustedBlockTimecode(ClusterTimecode, this.TimecodeScale) / TimecodeScale;
        frame.Duration = BlockDuration;
        frame.Reference = BlockReference;
        frame.Data = block.getFrame(0);
        FrameQueue.addLast(new MatroskaFrame(frame));

        if (block.getFrameCount() > 1) {
          for (int f = 1; f < block.getFrameCount(); f++) {
            frame.Data = block.getFrame(f);

            if (badMP3Headers()) {
              throw new RuntimeException("Bad Data!");
            }
            FrameQueue.addLast(new MatroskaFrame(frame));
            if (badMP3Headers()) {
              throw new RuntimeException("Bad Data!");
            }
          }
        }
      }

      level2.skipData(ioDS);
      level2 = ((MasterElement)level1).readNextChild(reader);
    }
  }

  protected boolean badMP3Headers() {
    ListIterator iter = FrameQueue.listIterator();
    while (iter.hasNext()) {
      MatroskaFrame frame = (MatroskaFrame)iter.next();
      if (frame.TrackNo == 2
          && frame.Data[3] != 0x54)
      {
        throw new RuntimeException("Bad MP3 Header! Index: " + iter.nextIndex());
      }
    }
    return false;
  }
  private void _parseSegmentInfo(Element level1, Element level2) {
    level2 = ((MasterElement)level1).readNextChild(reader);

    while (level2 != null) {
      if (level2.equals(MatroskaDocType.Title_Id)) {
        level2.readData(ioDS);
        SegmentTitle = ((StringElement)level2).getValue();

      } else if (level2.equals(MatroskaDocType.MuxingApp_Id)) {
        level2.readData(ioDS);
        MuxingApp = ((StringElement)level2).getValue();

      } else if (level2.equals(MatroskaDocType.WritingApp_Id)) {
        level2.readData(ioDS);
        WritingApp = ((StringElement)level2).getValue();

      } else if (level2.equals(MatroskaDocType.Duration_Id)) {
        level2.readData(ioDS);
        Duration = ((FloatElement)level2).getValue();

      } else if (level2.equals(MatroskaDocType.TimecodeScale_Id)) {
        level2.readData(ioDS);
        TimecodeScale = ((UnsignedInteger)level2).getValue();
      }

      level2.skipData(ioDS);
      level2 = ((MasterElement)level1).readNextChild(reader);
    }
  }

  private void _parseTracks(Element level1, Element level2) {
    Element level3 = null;
    Element level4 = null;
    level2 = ((MasterElement)level1).readNextChild(reader);

    while (level2 != null) {
      if (level2.equals(MatroskaDocType.TrackEntry_Id)) {
        MatroskaFileTrack track = new MatroskaFileTrack();
        level3 = ((MasterElement)level2).readNextChild(reader);

        while (level3 != null) {
          if (level3.equals(MatroskaDocType.TrackNumber_Id)) {
            level3.readData(ioDS);
            track.TrackNo = (short)((UnsignedInteger)level3).getValue();

          } else if (level3.equals(MatroskaDocType.TrackUID_Id)) {
            level3.readData(ioDS);
            track.TrackUID = ((UnsignedInteger)level3).getValue();

          } else if (level3.equals(MatroskaDocType.TrackType_Id)) {
            level3.readData(ioDS);
            track.TrackType = (byte)((UnsignedInteger)level3).getValue();

          } else if (level3.equals(MatroskaDocType.TrackDefaultDuration_Id)) {
            level3.readData(ioDS);
            track.DefaultDuration = ((UnsignedInteger)level3).getValue();

          } else if (level3.equals(MatroskaDocType.TrackName_Id)) {
            level3.readData(ioDS);
            track.Name = ((StringElement)level3).getValue();

          } else if (level3.equals(MatroskaDocType.TrackLanguage_Id)) {
            level3.readData(ioDS);
            track.Language = ((StringElement)level3).getValue();

          } else if (level3.equals(MatroskaDocType.TrackCodecID_Id)) {
            level3.readData(ioDS);
            track.CodecID = ((StringElement)level3).getValue();

          } else if (level3.equals(MatroskaDocType.TrackCodecPrivate_Id)) {
            level3.readData(ioDS);
            track.CodecPrivate = ((BinaryElement)level3).getData();

          } else if (level3.equals(MatroskaDocType.TrackVideo_Id)) {
            level4 = ((MasterElement)level3).readNextChild(reader);

            while (level4 != null) {
              if (level4.equals(MatroskaDocType.PixelWidth_Id)) {
                level4.readData(ioDS);
                track.Video_PixelWidth = (short)((UnsignedInteger)level4).getValue();

              } else if (level4.equals(MatroskaDocType.PixelHeight_Id)) {
                level4.readData(ioDS);
                track.Video_PixelHeight = (short)((UnsignedInteger)level4).getValue();

              } else if (level4.equals(MatroskaDocType.DisplayWidth_Id)) {
                level4.readData(ioDS);
                track.Video_DisplayWidth = (short)((UnsignedInteger)level4).getValue();

              } else if (level4.equals(MatroskaDocType.DisplayHeight_Id)) {
                level4.readData(ioDS);
                track.Video_DisplayHeight = (short)((UnsignedInteger)level4).getValue();
              }

              level4.skipData(ioDS);
              level4 = ((MasterElement)level3).readNextChild(reader);
            }

          } else if (level3.equals(MatroskaDocType.TrackAudio_Id)) {
            level4 = ((MasterElement)level3).readNextChild(reader);

            while (level4 != null) {
              if (level4.equals(MatroskaDocType.SamplingFrequency_Id)) {
                level4.readData(ioDS);
                track.Audio_SamplingFrequency = (float)((FloatElement)level4).getValue();

              } else if (level4.equals(MatroskaDocType.OutputSamplingFrequency_Id)) {
                level4.readData(ioDS);
                track.Audio_OutputSamplingFrequency = (float)((FloatElement)level4).getValue();

              } else if (level4.equals(MatroskaDocType.Channels_Id)) {
                level4.readData(ioDS);
                track.Audio_Channels = (short)((UnsignedInteger)level4).getValue();

              } else if (level4.equals(MatroskaDocType.BitDepth_Id)) {
                level4.readData(ioDS);
                track.Audio_BitDepth = (byte)((UnsignedInteger)level4).getValue();
              }

              level4.skipData(ioDS);
              level4 = ((MasterElement)level3).readNextChild(reader);
            }

          }
          level3.skipData(ioDS);
          level3 = ((MasterElement)level2).readNextChild(reader);
        }
        TrackList.add(track);
      }

      level2.skipData(ioDS);
      level2 = ((MasterElement)level1).readNextChild(reader);
    }
  }
  private void _parseTags(Element level1, Element level2) {
    Element level3 = null;
    Element level4 = null;
    level2 = ((MasterElement)level1).readNextChild(reader);

    while (level2 != null) {
      if (level2.equals(MatroskaDocType.Tag_Id)) {
        MatroskaFileTagEntry tag = new MatroskaFileTagEntry();
        level3 = ((MasterElement)level2).readNextChild(reader);

        while (level3 != null) {
          if (level3.equals(MatroskaDocType.TagTargets_Id)) {
            level4 = ((MasterElement)level3).readNextChild(reader);

            while (level4 != null) {
              if (level4.equals(MatroskaDocType.TagTargetTrackUID_Id)) {
                level4.readData(ioDS);
                tag.TrackUID.add(new Long(((UnsignedInteger)level4).getValue()));

              } else if (level4.equals(MatroskaDocType.TagTargetChapterUID_Id)) {
                level4.readData(ioDS);
                tag.ChapterUID.add(new Long(((UnsignedInteger)level4).getValue()));

              } else if (level4.equals(MatroskaDocType.TagTargetAttachmentUID_Id)) {
                level4.readData(ioDS);
                tag.AttachmentUID.add(new Long(((UnsignedInteger)level4).getValue()));
              }

              level4.skipData(ioDS);
              level4 = ((MasterElement)level3).readNextChild(reader);
            }

          } else if (level3.equals(MatroskaDocType.TagSimpleTag_Id)) {
            tag.SimpleTags.add(_parseTagsSimpleTag(level3, level4));
          }
          level3.skipData(ioDS);
          level3 = ((MasterElement)level2).readNextChild(reader);
        }
        TagList.add(tag);
      }

      level2.skipData(ioDS);
      level2 = ((MasterElement)level1).readNextChild(reader);
    }
  }

  private MatroskaFileSimpleTag _parseTagsSimpleTag(Element level3, Element level4) {
    MatroskaFileSimpleTag SimpleTag = new MatroskaFileSimpleTag();
    level4 = ((MasterElement)level3).readNextChild(reader);

    while (level4 != null) {
      if (level4.equals(MatroskaDocType.TagSimpleTagName_Id)) {
        level4.readData(ioDS);
        SimpleTag.Name = ((StringElement)level4).getValue();

      } else if (level4.equals(MatroskaDocType.TagSimpleTagString_Id)) {
        level4.readData(ioDS);
        SimpleTag.Value = ((StringElement)level4).getValue();

      } else if (level4.equals(MatroskaDocType.TagSimpleTag_Id)) {
        SimpleTag.Children.add(_parseTagsSimpleTag(level3, level4));
      }

      level4.skipData(ioDS);
      level4 = ((MasterElement)level3).readNextChild(reader);
    }

    return SimpleTag;
  }

  /**
   * Get a String report for the Matroska file.
   * Call readFile() before this method, else the report will be empty.
   *
   * @return String Report
   */
  public String getReport() {
    java.io.StringWriter s = new java.io.StringWriter();
    int t;

    s.write("MatroskaFile report\n");

    s.write("Infomation Segment \n");
    s.write("\tSegment Title: " + SegmentTitle + "\n");
    s.write("\tMuxingApp : " + MuxingApp + "\n");
    s.write("\tWritingApp : " + WritingApp + "\n");
    s.write("\tDuration : " + (int)(Duration/1000)/60 + "min \n");
    s.write("\tTimecodeScale : " + TimecodeScale + "\n");

    s.write("Track Count: " + TrackList.size() + "\n");
    for (t = 0; t < TrackList.size(); t++) {
      s.write("\tTrack " + t + "\n");
      s.write(TrackList.get(t).toString());
    }

    s.write("Tag Count: " + TagList.size() + "\n");
    for (t = 0; t < TagList.size(); t++) {
      s.write("\tTag Entry \n");
      s.write(TagList.get(t).toString());
    }

    s.write("End report\n");

    return s.getBuffer().toString();
  }
  
  public void setData() {
    for (int t = 0; t < TrackList.size(); t++) {
		MatroskaFileTrack track = (MatroskaFileTrack)TrackList.get(t);
		if (track.TrackType == MatroskaDocType.track_video) {
			VidCodec=track.CodecID;
			Resolution=track.Video_DisplayWidth + "x" + track.Video_DisplayHeight;
		}
		if (track.TrackType == MatroskaDocType.track_audio) {
			String tmp = track.CodecID;
			int i=tmp.indexOf("/");
	  		if(i!=-1) tmp=tmp.substring(0,i);
			AudCodec=AudCodec+tmp+",";
			
			Language=Language+track.Language+",";
			AudChannels=AudChannels+track.Audio_Channels+",";
			if (track.Audio_SamplingFrequency != 0)
				AudRate=AudRate+track.Audio_SamplingFrequency/1000+"KHz,";
		}
		if (track.TrackType == MatroskaDocType.track_subtitle) {
			if(!track.Language.equals("mis"))
				SubLang=SubLang+track.Language+",";
		}
    }	  
  }
  public String getSubLang() {
	  int i=SubLang.lastIndexOf(",");
	  if(i!=-1) SubLang=SubLang.substring(0,i);
	  return SubLang;
  }
  public String getVidCodec() {
	  return VidCodec;
  }
  public String getResolution() {
	  return Resolution;
  }
  public String getAudCodec() {
		int i=AudCodec.lastIndexOf(",");
        if(i!=-1)
            AudCodec=AudCodec.substring(0,i);
	  return AudCodec;
  }
  public String getAudRate() {
        int i=AudRate.lastIndexOf(",");
        if(i!=-1)
            AudRate=AudRate.substring(0,i);	  
	  return AudRate;
  }
  public String getLanguage() {
	  int i=Language.lastIndexOf(",");
	  if(i!=-1) Language=Language.substring(0,i);	  
	  return Language;
  }
  public void setAudChannels(String i) {
	  AudChannels=i;
  }
  public String getAudChannels() {
        int i=AudChannels.lastIndexOf(",");
        if(i!=-1)
            AudChannels=AudChannels.substring(0,i);		  
	  return AudChannels;
  }

  public String getWritingApp() {
    return WritingApp;
  }
  /**
   * Returns an array of the tracks
   */
  public MatroskaFileTrack [] getTrackList() {
    if (TrackList.size() > 0) {
      MatroskaFileTrack [] tracks = new MatroskaFileTrack[TrackList.size()];
      for (int t = 0; t < TrackList.size(); t++) {
        tracks[t] = (MatroskaFileTrack)TrackList.get(t);
      }
      return tracks;
    } else {
     return new MatroskaFileTrack[0];
    }
  }
  /**
   * This differs from the getTrackList method in that this method scans
   * each track and returns the one that has the same track number as TrackNo
   *
   * TrackNo != track index
   */
  public MatroskaFileTrack getTrack(int TrackNo) {
    for (int t = 0; t < TrackList.size(); t++) {
      MatroskaFileTrack track = (MatroskaFileTrack)TrackList.get(t);
      if (track.TrackNo == TrackNo)
        return track;
    }
    return null;
  }
  public long getTimecodeScale() {
    return TimecodeScale;
  }
  public String getSegmentTitle() {
    return SegmentTitle;
  }
  public String getMuxingApp() {
    return MuxingApp;
  }
  public int getDuration() {
    return (int)(Duration/1000);
  }
}
