/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.ext.util;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.io.File;
import java.io.RandomAccessFile;
import java.net.URL;
import java.util.Calendar;
import java.util.Random;

import javax.swing.ImageIcon;

import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.Repository;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DcResources;

/**
 * General utilities
 *
 * @author Robert Jan van der Waals
 * @since 1.3
 * @version 0.5
 */
public class Utilities {

    private static Random random = new Random();
    
    /**
     * Returns a centered location for a window / form / dialog 
     * @param windowSize size of the window
     * @return centered location
     */
    public static Dimension getCenteredWindowLocation(Dimension windowSize) {
        Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
        dim.height = (dim.height-windowSize.height)/2;
        dim.width = (dim.width-windowSize.width) /2;
        return dim;
    }
    
    public static String fillSubtitles(String filePath, String subtitles) {
        try {
            File f = new File(filePath);
            File dir = f.getParentFile();
            String movieName = f.getName().substring(0,f.getName().lastIndexOf("."));
            String [] allFiles = dir.list();
            for (int i = 0; i < allFiles.length; i++){
                String name = allFiles[i];
                if (    name.startsWith(movieName) && 
                        ( name.toUpperCase().endsWith("SRT") ||
                          name.toUpperCase().endsWith("SUB") ||
                          name.toUpperCase().endsWith("SSA"))) {
                    
                    if (!subtitles.equals("")) {
                        subtitles =subtitles + " ";
                    }
                    subtitles = subtitles + name.substring(movieName.length());
                }
            }
        } catch (Exception exp) {
            Messages.add(exp);
        }
        
        return subtitles;
    }
    
    private static final Calendar calendar = Calendar.getInstance();

    /**
     * Creates a unique ID. Can be used for custom IDs in the database.
     * Based on date / time + random number
     * @return unique ID as String
     */
    public static String getUniqueID() {
        long temp = random.nextLong();
        if (temp < 0l) {
            temp = temp * -1;
        }
        
        StringBuffer sb = new StringBuffer();
        sb.append(calendar.get(Calendar.MONTH) + 1);
        sb.append(calendar.get(Calendar.DAY_OF_MONTH));
        sb.append(temp);
        
        return sb.toString();
    }
    
    /**
     * Retrieved the file extension of a file
     * @param f file to get the extension from
     * @return extenstion or empty string
     */
    public static String getExtension(File f) {
        String name = f.getName().toLowerCase();
        int i = name.lastIndexOf( "." );
        if (i == -1) {
            return "";
        }
        return name.substring( i + 1 );
    }    
    
    public static int getIntegerValue(String s) {
        char[] characters = s.toCharArray();
        String test = "";
        for (int i = 0; i < characters.length; i++) {
            if (Character.isDigit(characters[i])) test += "" + characters[i];
        }
        
        int number = 0;
        try {
            number = Integer.valueOf(test).intValue();
        } catch (Exception ignore) {}
        
        return number;
    }
    
    
    /**
     * Retrieves the content of the file (first 100 bytes) so it can be 
     * compared with the content of another file
     * @param url url of the file to create the compare string for
     * @return string (content of file)
     * @throws Exception
     */
    public static String getCompareStringForFile(URL url) throws Exception{
        RandomAccessFile access = new RandomAccessFile(url.getFile(), "rw");
        byte[] byties = new byte[100];
        access.readFully(byties);
        access.close();
        String s = String.valueOf(Base64.encode(byties));
        return s.substring(0, s.length() - 2);
    }

    /**
     * Reads the content of a file (fully)
     * @param file file to retrieve the content from
     * @return content of the file as a byte array
     * @throws Exception
     */
    public static byte[] readFile(File file) throws Exception {
        RandomAccessFile access = new RandomAccessFile(file, "rw");
        byte[] byties = new byte[new Long(access.length()).intValue()];
        access.readFully(byties);
        access.close();
        return byties;        
    }

    public static ImageIcon base64ToImage(String base64) {
        byte[] bytes = Base64.decode(base64.toCharArray());
        return new ImageIcon(bytes);
    }
    
    public static String getHexColor(Color color) {
        String hexColor = "#" + Integer.toHexString(color.getRed());
        hexColor += Integer.toHexString(color.getGreen());
        hexColor += Integer.toHexString(color.getBlue()); 
        return hexColor.toUpperCase();
    }    
    
    public static ImageIcon getRatingIcon(Object value) {
        ImageIcon icon = IconLibrary._icoRating0;
        try {
            int rating = ((Integer) value).intValue();
            switch (rating) {
                case 1:
                    icon = IconLibrary._icoRating1;
                    break;
                case 2:
                    icon = IconLibrary._icoRating2;
                    break;
                case 3:
                    icon = IconLibrary._icoRating3;
                    break;
                case 4:
                    icon = IconLibrary._icoRating4;
                    break;                    
                case 5:
                    icon = IconLibrary._icoRating5;
                    break;                    
            }
        } catch (Exception ignore) {}
        return icon;
    }    

    public static String getCurrentDirectory() throws Exception {
    	File fl = new File(".");
    	fl = fl.getCanonicalFile();
        return fl.toString();
    }
    
    /**
     * Gets the content of a file and converts it to a base64 string
     * @param url url of file
     * @return base64 content of the file
     */
    public static String fileToBase64String(URL url) {
        try {
            if (url != null) {
                File file = new File(url.getFile());
                byte[] b = Utilities.readFile(file);
                file = null;
                return String.valueOf(Base64.encode(b));
            }
        } catch (Exception exp) {
            Messages.add(DcResources.getText(Repository.messages.msgCouldNotLoadPicture),
            		            Messages._ERROR);
            Messages.add(exp);
        }
        return "";
    }
}
